// file: $pdsp/src/v1.0/class/fourier_transform/fourier_transform.h
//
// this file defines the freq_response  class
//

// make sure definitions are only made once
//
#ifndef __FOURIER_TRANSFORM
#define __FOURIER_TRANSFORM

// isip include files
//
#ifndef __ISIP_INTEGRAL
#include <math/integral/integral.h>
#endif

// include system files
//
#include <time.h>

// Fourier_transform: a class that computes the fourier transform of
// some given data.
//
class Fourier_transform {
  
  //---------------------------------------------------------------------------
  //
  // protected data
  //
  //---------------------------------------------------------------------------
protected:
  
  // parameters related to the algorithm specification
  //
  int_4 N_d;					// order
  char_1* algorithm_name_d;			// specific algorithm
  int_4 algorithm_d;

  // parameters related to the signal data type
  //
  int_4 data_type_d;				// real or complex
  int_4 transform_direction_d;			// forward or reverse
  
  // parameters related to timing analysis
  //
  int_4 num_mults_d;				// multiplication counter
  int_4 num_adds_d;				// addition counter
  float_4 cpu_time_d;				// cpu time
  int_4 num_iter_d;				// number of iterations
  
  // debugging parameters
  //
  int_4 debug_level_d;				// amount of debugging info
  int_4 stats_d;                                // amount of statistics

  // some QFT related variables
  //
  float_4  **ws;//	Workspace pointers	
  float_4  *sc;	//	Secant table addr.	
  int_4	*nc;	//	Current QFT length	
  int_4	*ic;	//	Current secant inc.	
  int_4	*mc;	//	Current output pruning	
  int_4	nn;	//	Pruned QFT input length	
  int_4	mm;	//	Pruned QFT output length	
  int_4	ii;	//	Current QFT recursion level	


  //---------------------------------------------------------------------------
  //
  // public methods
  //
  //---------------------------------------------------------------------------
public:

  // required methods
  //
  char_1* name_cc();
  int_4 error_handler_cc(char_1* method_name, char_1* message);
  logical_1 get_params_cc(int argc, char** argv, int_4& order_a,
			  char_1* data_type_a, int_4& num_iter_a,
			  int_4& dbg_level_a, int_4& stats_a,
			  char_1* algo_name_a);
  
  // destructors/constructors
  //
  ~Fourier_transform();  
  Fourier_transform();
  
  // algorithm specification
  //
  logical_1 set_cc(int_4 order_a, int_4 type_a, int_4 num_iter_a,
		   int_4 dbg_level_a, int_4 stats_a, char_1* name_a);
  logical_1 method_choice_cc();
  
  // computation methods
  //
  logical_1 compute_cc(float_4* output, float_4* input);
  
  // timing methods
  //
  logical_1 reset_cc();
  logical_1 get_statistics_cc(int_4 num_mult, int_4 num_add,
			      int_4 mem_size, float_4 num_secs);
  logical_1 print_statistics_cc(FILE* fp, int_4 stats_level_a,
				int_4 num_iter_a);
  logical_1 set_cputime_cc(float_4 cpu_time_a);
  float_4 get_cputime_cc();
  
  // debugging methods
  //
  logical_1 set_debug_cc(int_4 debug_level);
  logical_1 debug_cc(FILE* fp, char_1* message);
  
  //---------------------------------------------------------------------------
  //
  // private methods
  //
  //---------------------------------------------------------------------------
private:

  // algorithm methods for real signals
  //
  logical_1 dft_real_cc(float_4* in_r, float_4* in_i);     // real DFT
  logical_1 fft2_real_cc(float_4* in_r, float_4* in_i);    // real radix-2
  //logical_1 qft_real_cc(float_4* real_coeff_a, float_4* in_r_a); // real QFT
  logical_1 fht_real_cc(float_4* in_r);                    // real FHT
  logical_1 srfft_real_cc(float_4* in_r, float_4* in_i);    // real SRFFT
  
  // algorithm methods for complex signals
  //
  logical_1 dft_complex_cc(float_4* in_r, float_4* in_i);  // complex DFT
  logical_1 fft2_complex_cc(float_4* in_r, float_4* in_i); // complex radix-2
  logical_1 fft4_complex_cc(float_4* in_r, float_4* in_i); // complex radix-4
  //  logical_1 qft_complex_cc(float_4* in_r, float_4* in_i);  // complex QFT
  logical_1 fht_complex_cc(float_4* in_r, float_4* in_i);  // complex FHT
  logical_1 srfft_complex_cc(float_4* in_r, float_4* in_i);// complex SRFFT

  // auxillary methods need for particular algorithms
  //
  logical_1 real_dct_cc(float_4* coeff_a,float_4* data_a); // DCT computation
  logical_1 real_dst_cc(float_4* coeff_a,float_4* data_a); // DST computation
  logical_1 qft_init_cc(); // initialize varibles for QFT computation
  logical_1 qft_close_cc(); // free memory after QFT computation
  int_4 powof2_cc(); // computes number of FFT stages required
  
  // lookup table generators 
  //
  logical_1 lookup_cc(float_4 *wreal, float_4 *wimag);
  logical_1 srfft_lookup_cc(float_4* cos_a, float_4* sin_a); // lookup table
  logical_1 srfft_new_lookup_cc(float_4* cos_a,
				float_4* sin_a); // lookup table
    
  // basic math operations (used to get timing information)
  //
  float_4 add_cc(float_4 x, float_4 y);
  float_4 sub_cc(float_4 x, float_4 y);
  float_4 mult_cc(float_4 x, float_4 y);
  float_4 mult_sum_cc(float_4 x, float_4 y, float_4 sum);

};

// end of include file
// 
#endif








