// file: $pdsp/src/v1.0/class/fourier_transform/ft_fht_0.cc
//

// isip include files
//
#include "fourier_transform.h"
#include "fourier_transform_constants.h"

// include file for the Fast hartley transform containing some
// necessary functions and lookup tables.
//
#define GOOD_TRIG
#include "trigtbl.h"

// method: fht_real_cc
//
// This code is based on the algorithm as described in the work by
// Ron Mayer whose code is available as a free-ware. The resulting array
// will have first half containing the real part  and the second half the
// imaginary part. This can be done because the fourier transform is
// symmetric.
//
// arguments:
//  float_4* real_a : data array
//
// return: a logical_1 value indicating status
//
// implements a real Fast Hartley transform
//
logical_1 Fourier_transform::fht_real_cc(float_4* real_a){
  
  float_4 a,b;
  float_4 c1,s1,s2,c2,s3,c3,s4,c4;
  float_4 f0,g0,f1,g1,f2,g2,f3,g3;
  int_4 i,k,k1,k2,k3,k4,kx;
  float_4 *fi,*fn,*gi;
  TRIG_VARS;

  // start timing process
  //
  cpu_time_d = clock();
  
  for (k1=1,k2=0;k1<N_d;k1++){
    for (k=N_d>>1; (!((k2^=k)&k)); k>>=1);
    if (k1>k2){
      a=real_a[k1];real_a[k1]=real_a[k2];real_a[k2]=a;
    }
  }
  for ( k=0 ; (1<<k)<N_d ; k++ ){};
  k  &= 1;
  if (k==0){
    for (fi=real_a,fn=real_a+N_d;fi<fn;fi+=4){
      f1 = sub_cc(fi[0 ],fi[1 ]);
      f0 = add_cc(fi[0 ],fi[1 ]);
      f3 = sub_cc(fi[2 ],fi[3 ]);
      f2 = add_cc(fi[2 ],fi[3 ]);
      fi[2] = sub_cc(f0,f2);	
      fi[0] = add_cc(f0,f2);
      fi[3] = sub_cc(f1,f3);	
      fi[1] = add_cc(f1,f3);
    }
  }
  else {
    
    for (fi=real_a,fn=real_a+N_d,gi=fi+1;fi<fn;fi+=8,gi+=8){
      c1 = sub_cc(fi[0 ],gi[0 ]);
      s1 = add_cc(fi[0 ],gi[0 ]);
      c2 = sub_cc(fi[2 ],gi[2 ]);
      s2 = add_cc(fi[2 ],gi[2 ]);
      c3 = sub_cc(fi[4 ],gi[4 ]);
      s3 = add_cc(fi[4 ],gi[4 ]);
      c4 = sub_cc(fi[6 ],gi[6 ]);
      s4 = add_cc(fi[6 ],gi[6 ]);
      f1 = sub_cc(s1,s2);	
      f0 = add_cc(s1,s2);
      g1 = sub_cc(c1,c2);	
      g0 = add_cc(c1,c2);
      f3 = sub_cc(s3,s4);	
      f2 = add_cc(s3,s4);
      g3 = mult_cc(SQRT2,c4);		
      g2 = mult_cc(SQRT2,c3);
      fi[4 ] = sub_cc(f0,f2);
      fi[0 ] = add_cc(f0,f2);
      fi[6 ] = sub_cc(f1,f3);
      fi[2 ] = add_cc(f1,f3);
      gi[4 ] = sub_cc(g0,g2);
      gi[0 ] = add_cc(g0,g2);
      gi[6 ] = sub_cc(g1,g3);
      gi[2 ] = add_cc(g1,g3);
    }
  }
  if (N_d<16) 

    // exit gracefully
    //
    return ISIP_TRUE;
  
  
  do
   {
     k  = (2 + k);
     k1 = 1 << k;
     k2 = k1 << 1;
     k4 = k2 << 1;
     k3 = k2+k1;
     kx = k1 >> 1;
     fi = real_a;
     gi = fi+kx;
     fn = real_a+N_d;
     do {
       f1 = sub_cc(fi[0 ],fi[k1]);
       f0 = add_cc(fi[0 ],fi[k1]);
       f3 = sub_cc(fi[k2],fi[k3]);
       f2 = add_cc(fi[k2],fi[k3]);
       fi[k2] = sub_cc(f0,f2);
       fi[0 ] = add_cc(f0,f2);
       fi[k3] = add_cc(f1,f3);
       fi[k1] = add_cc(f1,f3);
       g1 = sub_cc(gi[0 ],gi[k1]);
       g0 = add_cc(gi[0 ],gi[k1]);
       g3 = mult_cc(SQRT2,gi[k3]);
       g2 = mult_cc(SQRT2,gi[k2]);
       gi[k2] = sub_cc(g0,g2);
       gi[0 ] = add_cc(g0,g2);
       gi[k3] = sub_cc(g1,g3);
       gi[k1] = add_cc(g1,g3);
       gi = k4 + gi;
       fi = k4 + fi;
     } while (fi < fn);
     
     TRIG_INIT(k,c1,s1);
     for (i = 1; i < kx; i++){
       TRIG_NEXT(k,c1,s1);
       c2 = sub_cc(mult_cc(c1,c1),mult_cc(s1,s1));
       s2 = mult_cc(2,mult_cc(c1,s1));
       fn = (real_a + N_d);
       fi = (real_a + i);
       gi = (real_a + k1 - i);
       do{
	 b  = sub_cc(mult_cc(s2,fi[k1]),mult_cc(c2,gi[k1]));
	 a  = add_cc(mult_cc(c2,fi[k1]),mult_cc(s2,gi[k1]));
	 f1 = sub_cc(fi[0 ],a);
	 f0 = add_cc(fi[0 ],a);
	 g1 = sub_cc(gi[0 ],b);
	 g0 = add_cc(gi[0 ],b);
	 b  = sub_cc(mult_cc(s2,fi[k3]),mult_cc(c2,gi[k3]));
	 a  = add_cc(mult_cc(c2,fi[k3]),mult_cc(s2,gi[k3]));
	 f3 = sub_cc(fi[k2],a);
	 f2 = add_cc(fi[k2],a);
	 g3 = sub_cc(gi[k2],b);
	 g2 = add_cc(gi[k2],b);
	 b  = sub_cc(mult_cc(s1,f2),mult_cc(c1,g3));
	 a  = add_cc(mult_cc(c1,f2),mult_cc(s1,g3));
	 fi[k2]  = sub_cc(f0    ,a);
	 fi[0 ]  = add_cc(f0    ,a);
	 gi[k3]  = sub_cc(g1    ,b);
	 gi[k1]  = add_cc(g1    ,b);
	 b  = sub_cc(mult_cc(c1,g2),mult_cc(s1,f3));
	 a  = add_cc(mult_cc(s1,g2),mult_cc(c1,f3));
	 gi[k2]  = sub_cc(g0    ,a);
	 gi[0 ]  = add_cc(g0    ,a);
	 fi[k3]  = sub_cc(f1    ,b);
	 fi[k1]  = add_cc(f1    ,b);
	 gi = (gi + k4);
	 fi = (fi + k4);
       } while (fi < fn);
     }
     TRIG_RESET(k,c1,s1);
   } while (k4<N_d);

  // stop clock to get timing information
  //
  cpu_time_d = clock();
  
  // exit gracefully
  //
  return ISIP_TRUE;
}


