// file: $pdsp/src/v1.0/class/fourier_transform/ft_srfft_1.cc
//
 
// isip include files
//
#include "fourier_transform.h"
#include "fourier_transform_constants.h"
 
// method: srfft_complex_cc
//
// This code is based on the code by Sitton of the Rice University.
//
// arguments:
//  float_4* real_a : real part of input data
//  float_4* imag_a : imaginary part of input data
//
// return : logical_1 indicating no error in computation
//
// routine used to compute  the split-radix FFT
//
logical_1 Fourier_transform::srfft_complex_cc(float_4 *real_a, float_4 *imag_a)
{
  
  int_4	i, j, k, q, ii, i0, i1, i2, i3, n1, n2, n4, is, id;
  float_4 cc1, ss1, cc3, ss3, r1, r2, r3, s1, s2, t;
  float_4* cs = new float_4[N_d];
  float_4* ss = new float_4[N_d];
  
  // create sin-cos lookup table
  //
  srfft_lookup_cc(cs,ss);

  // bit reversal routine
  //
  n1 = N_d - 1;
  n2 = N_d >> 1;
  for (i = j = 0; i < n1; i++, j += k){
    if (i < j){
      t = real_a[j];
      real_a[j] = real_a[i];
      real_a[i] = t;
      t = imag_a[j];
      imag_a[j] = imag_a[i];
      imag_a[i] = t;
    }
    k = n2;
    while (k <= j){
      j -= k;
      k >>= 1;
    } 
  }
  
  // start timer
  //
  cpu_time_d = clock();
  
  // Length two butterflies	
  //
  int_4 m = powof2_cc();
  is = 0;
  id = 1;
  n1 = (int_4)sub_cc(N_d,1);
  
  do
    {
      id <<= 2;
      
      for (i0 = is; i0 < N_d; i0 += id){
	i1 = (int_4)add_cc(i0,1);

	r1 = real_a[i0];
	real_a[i0] = add_cc(r1,real_a[i1]);
	real_a[i1] = sub_cc(r1,real_a[i1]);
	
	r1 = imag_a[i0];
	imag_a[i0] = add_cc(r1,imag_a[i1]);
	imag_a[i1] = sub_cc(r1,imag_a[i1]); 
      }
      
      is = ((int_4)sub_cc(id,1)) << 1;
      
    } while (is < n1);
  
  // L shaped butterflies	
  //
  n2 = 2;
  
  for (k = 2; k <= m; k++){
    n4 = n2 >> 1;
    n2 <<= 1;
    
    is = 0;
    id = n2 << 1;
    
    // Unity twiddle factor loops	
    //
    do
      {
	for (i0 = is; i0 < n1; i0 += id){
	  i1 = (int_4)add_cc(i0,n4);
	  i2 = (int_4)add_cc(i1,n4);
	  i3 = (int_4)add_cc(i2,n4);
	  
	  r3 = add_cc(real_a[i2],real_a[i3]);
	  r2 = sub_cc(real_a[i2],real_a[i3]);
	  r1 = add_cc(imag_a[i2],imag_a[i3]);
	  s2 = sub_cc(imag_a[i2],imag_a[i3]);
	  
	  real_a[i2] = sub_cc(real_a[i0],r3);
	  real_a[i0] = add_cc(real_a[i0],r3);
	  real_a[i3] = sub_cc(real_a[i1],s2);
	  real_a[i1] = add_cc(s2,real_a[i1]);
	  
	  imag_a[i2] = sub_cc(imag_a[i0],r1);
	  imag_a[i0] = add_cc(imag_a[i0],r1);
	  imag_a[i3] = add_cc(imag_a[i1],r2);
	  imag_a[i1] = sub_cc(imag_a[i1],r2);
	  
	} 
	
	is = (id << 1) - n2;
	id <<= 2;
	
      } while (is < n1);
    
    q = ii = (int_4)(N_d / (float_4)n2);
    
    // Non-trivial twiddle factor loops	
    //
    for (j = 1; j < n4; j++, q += ii){
      cc1 = cs[q];
      ss1 = ss[q];
      i3 = (int_4)mult_cc((float_4)q,3.0);
      cc3 = cs[i3];
      ss3 = ss[i3];
      
      is = j;
      id = n2 << 1;
      
      do
	{
	  for (i0 = is; i0 < n1; i0 += id){
	    i1 = (int_4)add_cc(i0,n4);
	    i2 = (int_4)add_cc(i1,n4);
	    i3 = (int_4)add_cc(i2,n4);
	    
	    r1 = add_cc(mult_cc(real_a[i2],cc1),mult_cc(imag_a[i2],ss1));
	    s1 = sub_cc(mult_cc(imag_a[i2],cc1),mult_cc(real_a[i2],ss1));
	    r2 = add_cc(mult_cc(real_a[i3],cc3),mult_cc(imag_a[i3],ss3));
	    s2 = sub_cc(mult_cc(imag_a[i3],cc3),mult_cc(real_a[i3],ss3));
	    
	    r3 = add_cc(r1,r2);
	    r2 = sub_cc(r1,r2);
	    r1 = add_cc(s1,s2);
	    s2 = sub_cc(s1,s2);
	    
	    real_a[i2] = sub_cc(real_a[i0],r3);
	    real_a[i0] = add_cc(r3,real_a[i0]);
	    real_a[i3] = sub_cc(real_a[i1],s2);
	    real_a[i1] = add_cc(s2,real_a[i1]);
	    
	    imag_a[i2] = sub_cc(imag_a[i0],r1);
	    imag_a[i0] = add_cc(r1,imag_a[i0]);
	    imag_a[i3] = add_cc(imag_a[i1],r2);
	    imag_a[i1] = sub_cc(imag_a[i1],r2);
	  } 
	  
	  is = (int_4)add_cc(sub_cc((id << 1),n2),j);
	  id <<= 2;
	  
	} while (is < n1);
    }
  }
  
  // stop timer
  //
  cpu_time_d = clock();

  // free memory
  //
  delete cs,ss;
  
  // exit gracefully
  //
  return ISIP_TRUE;
}


