// file: $PDSP/class/fourier_transform/v1.1/ft_ditf_0.cc
//

// include files
//
#include "fourier_transform.h"
#include "fourier_transform_constants.h"

// define the core term of W
//
#define ARG_CORE ISIP_TWOPI/N_d

// DITF externals
//
static int_4 switch_stage;      // stage at which algorithm switches from dif
                                // to dit
static int_4* two_v_table;	// holds the powers of 2
static int_4** mod_table;	// holds the mod factors
static int_4** lambda_table;	// holds the possible values of lambda
				// see algorithms below
static float_8** dit_cosine_table;	// holds the cosine terms
static float_8** dit_sine_table;	// holds the sine terms
static float_8** ditf_cosine_table;	// holds the cosine terms
static float_8** ditf_sine_table;	// holds the sine terms
static float_8** convert_cosine_table;	// holds the cosine terms
static float_8** convert_sine_table;	// holds the sine terms

// method: ditf_real_cc
//
// This method is based on the DITF algorithm by Ali Saidi
// found in the 1994 ICASSP Proceedings 
//
// arguments:
// float_8* output_a :(output) output data array 
//                      length of output array will always be 2*N
//                      memory is allocated internally, not by the calling
//                      program.
// float_8* input_a : (input) input data array
//                      length of input array is N
//                      memory is allocated by the calling
//                      program.
//
// return: logical_1 indicating status
//
// this method performs the real ditf algorithm
//
logical_1 Fourier_transform::ditf_real_cc(float_8* output_a,
					  float_8* input_a) {

  // declare local variables
  //
  float_8 real_a[N_d];
  int_4 v = powof2_cc();
  int_4 i;			  // incremental counter index

  switch_stage = shift_cc(FT_RIGHT,v,1);

  // calculate 0..N/2 ditf coefficients
  //
  for (i=0;i<=(N_d/2);i++) {
    ditf_calc_cc(input_a,v,i,i,&real_a[i],&output_a[i]);
  }

  output_a[N_d/2] = 0;
  
  // Use the fft's symmetry to calculate the remaining coefficients.
  //
  for(i=1;((N_d/2)+i)<(N_d);i++) {
    real_a[(N_d/2)+i] = real_a[(N_d/2)-i];
    output_a[(N_d/2)+i] = -output_a[(N_d/2)-i];
  }

  // interlace output and put into output_a array
  //
   for ( int_4 kk = N_d - 1; kk >= 0; kk = sub_cc(kk,1) ){
    output_a[2*kk] = real_a[kk];
    output_a[2*kk+1] = output_a[kk];
  }
   output_a[1] = 0;
   
  // exit gracefully
  //
  return ISIP_TRUE;
  
}


// subroutine to calculate the dft using the recursive decimation-in-time
// algorithm described by Ali Saidi in the ICASSP Proceedings of 1994.
// this DIT should not be used alone because it outputs values in a special
// form to be consistent with the DITF
//
// input: input - array of sample points
//   		 v - stage currently in
//			 gamma - node of the stage
//			 two_v - 2^v
//			 n - sample point currently calculating
//			 N - number of samples
//			 real_result_ptr - pointer to the real part of the
//                            calculated result
//			 imag_result_ptr - pointer to the imag part of the
//                            calculated result
//
// output: None
//
// description: this function uses a recursive call to itself to calculate the
//		 coefficients of the dft at gamma node at stage v for the nth
//		 sample point using the following recursion.
//              if gamma = v = 0
//		       g'[gamma,v](n) = x(n)
//              if gamma < 2^(v-1)
//     	         g'[gamma,v](n) = g'[gamma,v-1](n) +
//		  g'[gamma,v-1]((n+N/(2^v))%N_d)W^(gamma/2^(v-1))
//		else
//		  lambda = gamma - 2^(v-1)
//		  g'[gamma,v](n) = g'[gamma,v-1](n) -
//		  g'[gamma,v-1]((n+N/(2^v))%N_d)W^(lambda/2^(v-1))
//
//
logical_1 Fourier_transform::ditf_dit_cc(float_8* input,int_4 v,int_4 gamma,
					 int_4 n,
					 float_8* real_result_ptr,
					 float_8* imag_result_ptr){
  // define local variables
  //
  float real1 = 0.0,
    real2 = 0.0,
    real3 = 0.0,
    imag1 = 0.0,
    imag2 = 0.0,
    imag3 = 0.0;
  float_8 x1=0;
  float_8 x2=0;
  int_4 half_two_v=0;
  
  // define variables that will be used often to avoid repeat calculations
  //
  int_4 v_minus = v-1;
  
  if (v != 0){
    half_two_v = two_v_table[v_minus];
  }
  
  // if v = gamma = 0
  // assign return x[n]
  //
  if ((v == 0) && (gamma == 0)) {
    *real_result_ptr = input[n];
    *imag_result_ptr = (float) 0;
  }
  
  // if gamma <= 2^(v-1) do recursion
  //
  else if (gamma < half_two_v) {
    
      x1 = (dit_cosine_table[gamma])[v];
      x2 = (dit_sine_table[gamma])[v];      
    
    // recursively compute the DIT
    //
    ditf_dit_cc(input,v_minus,gamma,n,&real1,&imag1);
    ditf_dit_cc(input,v_minus,gamma,(mod_table[n])[v],&real2,
		&imag2);
    
    // do complex multiplication and store results
    //
    if ((real2==0.0)&&(imag2==0.0)){
      real3=0;
      imag3=0;
    }
    else if ((real2==1.0)&&(imag2==1.0)) {
      real3 = add_cc(x1,x2);
      imag3 = sub_cc(x1,x2);
    }
    else if (x1==0.0) {
      if (x2==1.0){
	real3 = imag2;
	imag3 = sub_cc(0,real2);
      }
      else {
	real3 = sub_cc(0,imag2);
	imag3 = real2;
      }
    }
    else if(x2==0.0){
      if(x1==1.0){
	real3 = real2;
	imag3 = imag2;
      }
      else {
	real3 = -real2;
	imag3 = -imag2;
      }
    }
    else {

      real3 = add_cc(mult_cc(real2,x1),mult_cc(imag2,x2));
      imag3 = sub_cc(mult_cc(imag2,x1),mult_cc(real2,x2));
    }
    
    if (real3==0.0){
      *real_result_ptr = real1;
    }
    else{
      *real_result_ptr = add_cc(real1,real3);
    }
    if (imag3==0.0){
      *imag_result_ptr=imag1;
    }
    else{
      *imag_result_ptr = add_cc(imag1,imag3);
    }
  }    
  
  
  // if gamma > 2^(v-1) then do recursion
  //
  else {
    
    // assign lambda = gamma - 2^(v-1)
    //
    int lambda = (lambda_table[gamma])[v];
    
    // compute complex multiplier
    //
      x1 = (dit_cosine_table[lambda])[v];
      x2 = (dit_sine_table[lambda])[v];            
    
    // recursively compute DIT
    //
    ditf_dit_cc(input,v_minus,lambda,mod_table[n][0],&real1,&imag1);
    ditf_dit_cc(input,v_minus,lambda,mod_table[n][v],
		&real2,&imag2);
    
    // do complex multiplication and store results
    //
    if ((real2==0.0)&&(imag2==0.0)) {
      real3=0.0;
      imag3=0.0;
    }
    else if ((real2==1.0)&&(imag2==1.0)) {
      real3=add_cc(x1,x2);
      imag3=sub_cc(x1,x2);
    }
    else if (x1==0.0) {
      if (x2==1.0){
	real3 = imag2;
	imag3 = -real2;
      }
      else {
	real3 = -imag2;
	imag3 = real2;
      }
    }
    else if (x2==0.0) {
      if (x1==1.0){
	real3 = real2;
	imag3 = imag2;
      }
      else {
	real3 = -real2;
	imag3 = -imag2;
      }
    }
    
    else {
      real3 = add_cc(mult_cc(real2,x1),mult_cc(imag2,x2));
      imag3 = sub_cc(mult_cc(imag2,x1),mult_cc(real2,x2));
    }
    
    if (real3==0.0){
      *real_result_ptr = real1;
    }
    else{
      *real_result_ptr = sub_cc(real1,real3);
    }
    if (imag3==0.0){
      *imag_result_ptr=imag1;
    }
    else{
      *imag_result_ptr = sub_cc(imag1,imag3);
    }
  }
  
  // exit gracefully
  //
  return ISIP_TRUE;
}


// subroutine to calculate the dft using the decimation-in-time-frequency
// algorithm described by Ali Saidi in the ICASSP Proceedings of 1994.
//
// input: input - array of sample points
//   		 v - stage currently in
//			 gamma - node of the stage
//			 two_v - 2^v
//			 n - sample point currently calculating
//			 N - number of samples
//			 real_result_ptr - pointer to the real part of the
//                           calculated result
//			 imag_result_ptr - pointer to the imag part of the
//                           calculated result
//
// output: None
//
// description: this function uses a recursive call to itself to calculate the
// 	 coefficients of the dft at the gamma node of stage v for the nth
//	 sample point using the following recursion.
//         if gamma = v = 0
//		g[gamma,v](n) = x(n)
//         if gamma < 2^(v-1)
//		g[gamma,v](n) = g[gamma,v-1](n) +
//		 g[gamma,v-1]((n+N/(2^v))%N)
//	 else
//        	 lambda = gamma - 2^(v-1)
//	         g[gamma,v](n) = (g[gamma,v-1](n) -
//		  g[gamma,v-1]((n+N/(2^v))%N))W^(n*2^(v-1))
//
//	this is repeated until v = (log2(N))/2
//	at that point the remaining iterations are carried out by the
//	recursive decimation-in-time algorithm and the result is
//	multiplied by W^(gamma*n)
//
logical_1 Fourier_transform::ditf_calc_cc(float_8* input,int_4 v,
					  int_4 gamma,int_4 n,
					  float_8* real_result_ptr,
					  float_8* imag_result_ptr){
  // local variables
  //
  float_8 real1 = 0.0,
    real2 = 0.0,
    imag1 = 0.0,
    imag2 = 0.0;
  float_8 x1 = 0.0,
    x2 = 0.0;
  int_4 half_two_v=0;


  // determine if it is time to switch to the DIT portion of the recursion
  //
  if ( switch_stage < v ) {

    
    //  assign variables which are used multiple times to avoid recalculation
    //
    int v_minus = sub_cc(v,1);
    if (v != 0){
      half_two_v = two_v_table[v_minus];
    }

    // if v = gamma = 0, return x[n]
    //
    if ((v == 0) && (gamma == 0)) {
      *real_result_ptr = input[n];
      *imag_result_ptr = (float) 0;
    }

    // if gamma <= 2^(v-1) do recursion
    //
    else if (gamma < (half_two_v)) {
      ditf_calc_cc(input,v_minus,gamma,mod_table[n][0],&real1,
		   &imag1);
      ditf_calc_cc(input,v_minus,gamma,mod_table[n][v],
		   &real2,&imag2);

      if (real1==0.0) {
	*real_result_ptr = real2;
      }
      else if (real2==0.0) {
	*real_result_ptr = real1;
      }
      else {
	*real_result_ptr = add_cc(real1,real2);
      }
      if (imag1==0.0) {
	*imag_result_ptr = imag2;
      }
      else if(imag2==0.0){
	*imag_result_ptr=imag1;
      }
      else {
	*imag_result_ptr = add_cc(imag1,imag2);
      }
    }
    
    
    // if gamma > 2^(v-1) do recursion
    //
    else {
      int lambda = (lambda_table[gamma])[v];

      x1 = (ditf_cosine_table[n])[v];
      x2 = (ditf_sine_table[n])[v];
      
      ditf_calc_cc(input,v_minus,lambda,mod_table[n][0],&real1,
		   &imag1);
      ditf_calc_cc(input,v_minus,lambda,mod_table[n][v],
		   &real2,&imag2);
      
      if(real2!=0.0) {
	real1 = sub_cc(real1,real2);
      }
      if (imag2!=0.0) {
	imag1 = sub_cc(imag1,imag2);
      }
      
      
      // do complex multiplication
      //
      if ((real1==0.0)&&(imag1==0.0)) {
	*real_result_ptr = 0;
	*imag_result_ptr = 0;
      }
      else if (x1 == 0) {
	if (x2==1.0) {
	  *real_result_ptr = imag1;
	  *imag_result_ptr = -real1;
	}
	else {
	  *real_result_ptr = -imag1;
	  *imag_result_ptr = real1;
	}
      }
      else if (x2 == 0) {
	if (x1==1.0) {
	  *real_result_ptr = real1;
	  *imag_result_ptr = imag1;
	}
	else {
	  *real_result_ptr = -real1;
	  *imag_result_ptr = -imag1;
	}	
      }
      
      else {
	*real_result_ptr = add_cc(mult_cc(real1,x1),mult_cc(imag1,x2));
	*imag_result_ptr = sub_cc(mult_cc(imag1,x1),mult_cc(real1,x2));
      }
    }
  }
  
  // switch to the DIT portion of the recursion
  //
  else {
    x1 = (convert_cosine_table[n])[gamma];
    x2 = (convert_sine_table[n])[gamma];
    
    // call the DIT recursion and multiply by conversion factor
    //
    ditf_dit_cc(input,v,gamma,mod_table[n][0],&real1,&imag1);

    if ((imag1==0.0)&&(real1==0.0)){
      *real_result_ptr = 0;
      *imag_result_ptr = 0;
    }
    else if (x1 == 0) {
      if(x2==1) {
	*real_result_ptr = imag1;
	*imag_result_ptr = -real1;
      }
      else {
	*real_result_ptr = -imag1;
	*imag_result_ptr = real1;
      }	
    }
    else if (x2==0) {
      if(x1==1) {
	*real_result_ptr = real1;
	*imag_result_ptr = imag1;
      }
      else {
	*real_result_ptr = -real1;
	*imag_result_ptr = -imag1;
      }
    }
    else if ((imag1==1.0)||(real1==1.0)){
      if ((imag1==1)&&(real1==1)){
	*real_result_ptr = add_cc(x1,x2);
	*imag_result_ptr = sub_cc(x1,x2);
      }
      else if(imag1==1){
	*real_result_ptr = add_cc(mult_cc(real1,x1),x2);
	*imag_result_ptr = sub_cc(x1,mult_cc(real1,x2));
      }
      else if(real1==1){
	*real_result_ptr = add_cc(x1,mult_cc(imag1,x2));
	*imag_result_ptr = sub_cc(mult_cc(imag1,x1),x2);
      }
    }
    
    else {	
      *real_result_ptr = add_cc(mult_cc(real1,x1),mult_cc(imag1,x2));
      *imag_result_ptr = sub_cc(mult_cc(imag1,x1),mult_cc(real1,x2));
    }
    
  }

  // exit gracefully
  //
  return ISIP_TRUE;

}



// method: ditf_init_cc
//
// This method is based on the DITF algorithm by Ali Saidi
// found in the 1994 ICASSP Proceedings 
//
// return: a logical_1 value indicating status
//
logical_1 Fourier_transform::ditf_init_cc() {


  // declare local variables
  //
  
  int_4 v = powof2_cc();
  int_4 i,t;			  // incremental counter index
  int_4 two_v;  		  // two_v = 2^v
  int_4 k;
  float_8 arg;
  int_4 max_gamma=1;


  // allocate memory for one dimensional array of pointers
  // which will be used to create a two dimensional array
  // mod_table.
  //
  mod_table = new int_4*[N_d];

  
  // allocate memory for one dimensional array of pointers
  // which will be used to create a two dimensional array
  // of trig functions.  These functions are the conversion
  // factors between the dit and dif portions of the ditf.
  //
  convert_cosine_table = new float_8*[N_d];
  convert_sine_table = new float_8*[N_d];  
  

  // allocate memory for one dimensional array of pointers
  // which will be used to create a two dimensional array
  // lambda_table.
  //
  lambda_table = new int_4*[N_d];

  // allocate memory for one dimensional array of pointers
  // which will be used to create a two dimensional array
  // for the cosine table and sine tables needed by this
  // program
  //
  dit_cosine_table = new float_8*[N_d];
  dit_sine_table = new float_8*[N_d];
  ditf_cosine_table = new float_8*[N_d];
  ditf_sine_table = new float_8*[N_d];

  // allocate memory for a one dimensional array
  // holding the powers of two.
  //
  two_v_table = new int_4[v+1];

  // create lookup table for two_v coeff's
  // the cell at location two_v_table[i] = 2^i
  //
  two_v = 1;
  for (t=0;t<=v;t++){
    two_v_table[t]=two_v;
    two_v = two_v*2;
  }

  // create a variable used to set the lookup table size
  //
  for( i=1; i<v; i=i+2 ) {
    max_gamma = max_gamma + two_v_table[(i-1)/2];
  }
  
  // create lookup table for mod products
  // The cell at the location mod_table[i][j]
  // is equal to (i + N/(2^j)) mod N
  // The cell at the location lambda_table[i][j]
  // is equal to ( i - 2^(v-1) )
  // each of the cosine and sine tables are used as defined in
  // in the DIT and DITF algorithms
  //
  for(i=0;i<N_d;i++){

    // initialize memory pointers for each 2-d array
    //
    lambda_table[i] = new int_4[v+1];
    mod_table[i] = new int_4[v+1];
    dit_cosine_table[i] = new float_8[v+1];
    dit_sine_table[i] = new float_8[v+1];
    ditf_cosine_table[i] = new float_8[v+1];
    ditf_sine_table[i] = new float_8[v+1];
    convert_cosine_table[i] = new float_8[max_gamma];
    convert_sine_table[i] = new float_8[max_gamma];

    (mod_table[i])[0] = i;
    (lambda_table[i])[0] = 0;
    (dit_cosine_table[i])[0] = 0;
    (dit_sine_table[i])[0] = 0;
    (ditf_cosine_table[i])[0] = 0;
    (ditf_sine_table[i])[0] = 0;
    
    for(k=1;k<=v;k++){
      (mod_table[i])[k]=(i+N_d/two_v_table[k])%N_d;
      (lambda_table[i])[k] = (i-two_v_table[k-1]);
      arg = ARG_CORE * (float_8)(i)*two_v_table[k-1];
      (ditf_cosine_table[i])[k] = cos(arg);
      (ditf_sine_table[i])[k] = sin(arg);
      arg = ARG_CORE * (float_8)(i * N_d) / (float_8)two_v_table[k];
      (dit_cosine_table[i])[k] = cos(arg);
      (dit_sine_table[i])[k] = sin(arg);
    }
    for(k=0;k<max_gamma;k++) {
      (convert_cosine_table[i])[k] = cos(ARG_CORE*i*k);
      (convert_sine_table[i])[k] = sin(ARG_CORE*i*k);
    }
  }

  // exit gracefully
  //
  return ISIP_TRUE;
}

// method: ditf_cleanup_cc
//
// This method is based on the DITF algorithm by Ali Saidi
// found in the 1994 ICASSP Proceedings 
//
// return: a logical_1 value indicating status
//
// this method deallocates memory declared for the ditf computations
//
logical_1 Fourier_transform::ditf_cleanup_cc() {
  
  int i;
  for(i=0;i<N_d;i++){

    // free memory pointers for each 2-d array
    //
    delete [] lambda_table[i];
    delete [] mod_table[i];
    delete [] dit_cosine_table[i];
    delete [] dit_sine_table[i];
    delete [] ditf_cosine_table[i];
    delete [] ditf_sine_table[i];
    delete [] convert_cosine_table[i];
    delete [] convert_sine_table[i];
  }
  // free allocated memory
  //
  delete [] mod_table;
  delete [] two_v_table;
  delete [] lambda_table;
  delete [] dit_cosine_table;
  delete [] dit_sine_table; 
  delete [] ditf_cosine_table;
  delete [] ditf_sine_table; 
  delete [] convert_cosine_table;
  delete [] convert_sine_table;

  // exit gracefully
  //
  return ISIP_TRUE;  
}

