// file: $PDSP/class/fourier_transform/v1.1/ft_fft2_1.cc
//

// isip include files
//
#include "fourier_transform.h"
#include "fourier_transform_constants.h"

// method: fft2_complex_cc
//
/// this method performs a radix-2 cooley-tukey decimation-in-frequency
// algorithm.  the code follows the exact structure of that in the book by
// proakis and manolakis " digital signal processing " pp. 731-732
//
// arguments:
// float_8* output_a :(output) output data array 
//			length of output array will always be 2*N
//			memory is allocated internally, not by the calling
//			program.
// float_8* input_a : (input) input data array
//			length of input array is 2*N
//			input data memory should be allocated by the
//			calling program.
//
// return: a logical_1 value indicating status
//
// implements a complex discrete fourier transform
//
logical_1 Fourier_transform::fft2_complex_cc( float_8* output_a,
					      float_8* input_a){
  // declare local variables
  //
  int_4 m,l,i1,i2;
  int_4 n1,n2;
  float_8 real_twid,imag_twid;
  float_8 tempr, tempi; 
  float_8 temp_input[N_d];

  // get the lookup values
  //
  float_8 *wr = new float_8[N_d];
  float_8 *wi = new float_8[N_d];
   
  // create lookup table
  //
  lookup_cc((float_8*)wr,(float_8*)wi);

  // copy input imaginary data to temporary memory and input real data
  // into output so that the input data is retained
  // after calculations.
  //
  for (int_4 kk = 0; kk < N_d; kk++) {
    temp_input[kk] = input_a[2*kk+1];
    output_a[kk] = input_a[2*kk];
  }
  
  
  m = powof2_cc();
  n2 = N_d;
  
  for(int_4 i = 0; i < m; i = add_cc(i,1)){
    n1 = n2;
    n2 = shift_cc(FT_RIGHT,n2,1);
    i1 = 0;
    i2 = (int_4)((N_d)/(float_8)n1);
    
    for(int_4 j = 0; j < n2; j = add_cc(j,1)){
      real_twid = wr[i1];
      imag_twid = wi[i1];
      i1= add_cc(i1,i2);
      
      for(int_4 k = j; k < N_d; k = add_cc(k,n1)){
	l = add_cc(k,n2);
	tempr = sub_cc(output_a[k],output_a[l]);
	output_a[k] = add_cc(output_a[k],output_a[l]);
	tempi = sub_cc(temp_input[k],temp_input[l]);
	temp_input[k] = add_cc(temp_input[k],temp_input[l]);
	output_a[l] = add_cc(mult_cc(real_twid,tempr),
			     mult_cc(imag_twid,tempi));
	temp_input[l] = sub_cc(mult_cc(real_twid,tempi),
			       mult_cc(imag_twid,tempr));
      }
    }
  }

  //procedure for bit reversal
  //
  float_8 temp = 0;
  int j = 0;
  n1 = sub_cc(N_d,1);
  n2 = shift_cc(FT_RIGHT,N_d,1);
  
  for (int_4 i = 0; i < n1 ; ++i){
    if (i < j)
      {
	temp = output_a[j];
	output_a[j] = output_a[i];
	output_a[i] = temp;
	temp = temp_input[j];
	temp_input[j] = temp_input[i];
	temp_input[i] = temp;
      }
    int_4 k = n2;
    while (k <= j)
      {
	j = sub_cc(j,k);
	k = shift_cc(FT_RIGHT,k,1);
      }
    j = add_cc(j,k);
  }


  // interlace the output and store in the output_a array
  //
  for (int_4 kk = N_d - 1; kk >= 0; kk = sub_cc(kk,1)) {
    output_a[2*kk+1] = temp_input[kk];
    output_a[2*kk] = output_a[kk];
  }

  if ( ISIP_DEBUG_NONE < debug_level_d) {
    
    // diagnostic printing
    //
    for (int_4 kk = 0; kk < N_d; kk++) {
      printf("output[%ld] = %f    + j %f\n", kk, output_a[2*kk],
	     output_a[2*kk+1]);
    }
    
  }
  
  // delete memory
  //
  delete wr;
  delete wi;
  
  // exit gracefully
  //

  return ISIP_TRUE;
}

