// file: $PDSP/class/fourier_transform/v1.1/ft_fht_1.cc
//

// isip include files
//
#include "fourier_transform.h"
#include "fourier_transform_constants.h"

// method: fht_complex_cc
//
// This code is based on the algorithm as described in the work by
// Ron Mayer whose code is available as a free-ware. The algorithm is
// described in the paper Hsieh S. Hou,"The Fast Hartley Transform Algorithm",
// IEEE Transactions on Computers, February 1987.pp. 147-153
//
//
// arguments:
// float_8* output_a :(output) output data array 
//			length of output array will always be 2*N
//			memory is allocated internally, not by the calling
//			program.
// float_8* input_a : (input) input data array
//			length of input array is 2*N
//			input data memory should be allocated by the
//			calling program.
//
// return: a logical_1 value indicating status
//
// implements a complex Fast Hartley transform
//
logical_1 Fourier_transform::fht_complex_cc(float_8* output_a,
					    float_8* input_a) {

  // declare local variables
  //
  float_8 a,b,c,d;
  int_4 i,j,k;
  float_8 q,r,s,t;
  float_8 temp_input[2*N_d];
  float_8 temp_output_a[N_d];
  float_8 temp_input_a[N_d];
  
  // copy input data to  temporary memory and zero the output so that
  // the input data is retained after calculations.
  //
  for (int_4 kk = 0; kk < N_d; kk++) {
    temp_output_a[kk] = input_a[2*kk];
    temp_input_a[kk] = input_a[2*kk + 1];
  }

  // prepare data for FHT computations
  //
  for (i = 1,j = sub_cc(N_d,1), k = (N_d/2); i < k; i = add_cc(i,1),j = sub_cc(j,1)) {
    a = temp_output_a[i];
    b = temp_output_a[j];
    q = add_cc(a,b);
    r = sub_cc(a,b);
    c = temp_input_a[i];
    d = temp_input_a[j];
    s = add_cc(c,d);
    t = sub_cc(c,d);
    temp_output_a[i] = mult_cc(add_cc(q,t),0.5);
    temp_output_a[j] = mult_cc(sub_cc(q,t),0.5);
    temp_input_a[i] = mult_cc(sub_cc(s,r),0.5);
    temp_input_a[j] = mult_cc(add_cc(s,r),0.5);
  }
  
  // compute hartley transform of the real and imaginary parts separately
  //
  fht_real_cc(output_a,temp_output_a);
  fht_real_cc(temp_input,temp_input_a);
  
  // interlace output and put into output_a array
  //
  for ( int_4 kk = N_d - 1; kk >= 0; kk = sub_cc(kk,1) ){
    output_a[2*kk] = output_a[kk];
    output_a[2*kk+1] = temp_input[kk];
  }

  // zero the imaginary portion of the N_d/2 coefficient
  //
  output_a[N_d+1] = 0.0;
  
  if ( ISIP_DEBUG_NONE < debug_level_d) {
    
    // diagnostic printing
    //
    for (int_4 kk = 0; kk < N_d; kk++) {
      printf("output[%ld] = %f    + j %f\n", kk, output_a[2*kk],
	     output_a[2*kk+1]);
    }
    
  }
  
  // exit gracefully
  //
  return ISIP_TRUE;
}



