// file: $PDSP/class/fourier_transform/v1.1/ft_qft_0.cc
//

// system include files
//
#include <math.h>

// isip include files
//
#include "fourier_transform.h"
#include "ft_qft_0.h"
#include "fourier_transform_constants.h"

// This code is based on the algorithm described in
// H.Guo et.al,"The Quick Discrete Fourier Transform", IEEE Transactions
// on Acoustics Speech and Signal Processing 1994.pp. 445-448
// This code is adapted form the code sent by Sitton of Rice University
//
// arguments:
// float_8* output_a :(output) output data array 
//                      length of output array will always be 2*N
//                      memory is allocated internally, not by the calling
//                      program.
// float_8* input_a : (input) input data array
//                      length of input array is N
//                      memory is allocated by the calling
//                      program.
//
// return: logical_1 indicating status
//
// implements a real Quick Fourier Transform
//
logical_1 Fourier_transform::qft_real_cc( float_8* output_a,
					  float_8* input_a ) {
  // declare local variables
  //
  float_8 data_a[N_d];
  float_8* xx;
  float_8 real_coeff_a[N_d];
  float_8 imag_coeff_a[N_d];
  float_8* y;
  int_4	i, n2;

  // copy input to temporary memory so that the input data is retained
  // after calculations.
  //
  for (int_4 kk = 0; kk < N_d; kk++) {
    data_a[kk] = input_a[kk];
  }

  // initialize memory for qft computation
  //
  if (qft_init_cc() == ISIP_FALSE) {
    return ISIP_FALSE;
  }
  
  // Set up constants	
  //
  if (N_d < MM)
    nn = MM;
  else
    nn = N_d;
  
  mm = N_d;
  n2 = shift_cc(FT_RIGHT,N_d,1);
  xx = ws[ii = 0];
  
  for (i = 1; i < mc[0]; i = add_cc(i,1))
    mc[i] = N_d;

  // Form even function
  //
  if (nn <= n2) {
    y = data_a;
  }
  
  else {
    xx[0] = data_a[0];
    
    for (i = 1; i <= sub_cc(N_d,nn); i = add_cc(i,2)){
      xx[i] = data_a[i];
      xx[add_cc(i,1)] = data_a[add_cc(i,1)];
    }
    
    xx[i] = add_cc(data_a[i],data_a[sub_cc(N_d,i)]);
    i = add_cc(i,1);
    
    for (;i < n2; i = add_cc(i,2)) {
      xx[i] = add_cc(data_a[i],data_a[sub_cc(N_d,i)]);
      xx[i+1] = add_cc(data_a[add_cc(i,1)],
			       data_a[sub_cc(N_d,add_cc(i,1))]);
            }
          
          xx[i] = data_a[i];
	  i = add_cc(i,1);
 
          for (;i < MM; i = add_cc(i,1))
            xx[i] = 0.0;
          
          y = xx;
        }

  // Do a real DCT
  //
  real_dct_cc(real_coeff_a, y);

  //    Form negative odd function
  //
  if (nn <= n2) {
      xx[1] = -data_a[1];
      
      for (i = 2; i < nn; i = add_cc(i,2)) {
          xx[i] = -data_a[i];
          xx[i+1] = -data_a[add_cc(i,1)];
        }
  }
  else {
    
    for (i = 1; i <= sub_cc(N_d,nn); i = add_cc(i,2)) {
      xx[i] = -data_a[i];
      xx[i+1] = -data_a[add_cc(i,1)];
    }
    
    xx[i] = sub_cc(data_a[N_d-i],data_a[i]);
    i = add_cc(i,1);
      
      for (;i < n2; i = add_cc(i,2)) {
	xx[i] = sub_cc(data_a[N_d-i],data_a[i]);
	xx[i+1] = sub_cc(data_a[N_d-(i+1)],data_a[i+1]);
      }
  }
  
  for (; i < MM; i = add_cc(i,1))
    xx[i] = 0.0;
  
  // Do a real DST
  //
  real_dst_cc(&imag_coeff_a[1], &xx[1]);
  imag_coeff_a[0] = imag_coeff_a[n2] = 0.0;
  
  // create symmetric part of the spectrum
  //
  for (int j = 1;j <= N_d/2;j++){
    imag_coeff_a[N_d - j] = -imag_coeff_a[j];
    real_coeff_a[N_d - j] = real_coeff_a[j];
  }
  imag_coeff_a[N_d/2] = 0.0;
  
  // interlace output into output_a
  //
  for ( int_4 kk = N_d - 1; kk >= 0; kk = sub_cc(kk,1) ){
    output_a[2*kk] = real_coeff_a[kk];
    output_a[2*kk+1] = imag_coeff_a[kk];
  }

  // deallocate memory
  //
  qft_close_cc();

  // exit gracefully
  //
  return ISIP_TRUE;
}

//
// This code is based on the algorithm described in
// H.Guo et.al,"The Quick Discrete Fourier Transform", IEEE Transactions
// on Acoustics Speech and Signal Processing 1994.pp. 445-448
// This code is adapted form the code sent by Sitton of Rice University
//
// arguments:
// float_8* output_a :(output) output data array 
//                      length of output array will always be 2*N
//                      memory is allocated internally, not by the calling
//                      program.
// float_8* input_a : (input) input data array
//                      length of input array is 2*N
//                      memory is allocated by the calling
//                      program.
//
// return: a logical_1 value indicating status
//
// implements a complex quick fourier transform
//
logical_1 Fourier_transform::qft_complex_cc( float_8* output_a,
					     float_8* input_a ) {
  
  int_4	i, n2, n4;
  float_8 real_data_a[N_d];
  float_8 imag_data_a[N_d];
  float_8 real_coeff_a[2*N_d];
  float_8 imag_coeff_a[2*N_d];
  
  // copy input imaginary data to temporary memory and input real data
  // into output so that the input data is retained
  // after calculations.
  //
  for (int_4 kk = 0; kk < N_d; kk++) {
    real_data_a[kk] = input_a[2*kk];
    imag_data_a[kk] = input_a[2*kk+1];
  }
  
  n2 = shift_cc(FT_RIGHT,N_d, 1);
  n4 = shift_cc(FT_RIGHT,n2,1);
  
  // Compute packed QFT of the real and imaginary parts	
  //
  qft_real_cc(real_coeff_a,real_data_a);
  qft_real_cc(imag_coeff_a,imag_data_a);

  // interlace the output data into output_a
  //
  for (i = 0; i < N_d; i = add_cc(i,1)){
    output_a[2*i] = sub_cc(real_coeff_a[2*i],imag_coeff_a[2*i+1]);
    output_a[2*i+1]= add_cc(real_coeff_a[2*i+1],imag_coeff_a[2*i]);
  }

  if ( ISIP_DEBUG_NONE < debug_level_d) {
    
    // diagnostic printing
    //
    for (int_4 kk = 0; kk < N_d; kk++) {
      printf("output[%ld] = %f    + j %f\n", kk, output_a[2*kk],
	     output_a[2*kk+1]);
    }
    
  }
 
  // exit gracefully
  //
  return ISIP_TRUE;
} 

//
// This code is based on the algorithm described in
// H.Guo et.al,"The Quick Discrete Fourier Transform", IEEE Transactions
// on Acoustics Speech and Signal Processing 1994.pp. 445-448
// This code is adapted form the code sent by Sitton of Rice University
//
// arguments:
//  float_8* coeff_a : (output) output dct coefficient array
//  float_8* data_a  : (input) input data array
//
// return: a logical_1 value indicating status
//
// implements a disrete cosine transform
//
logical_1 Fourier_transform::real_dct_cc(float_8* coeff_a, float_8* data_a){
  
  float_8 t0, t1, t2, t3, t4;
  int_4	i, j, k, n, nq, n2, n2p1;
  
  float_8* XX;
  float_8* xx;
  
  // Increment level	& start	
  //
  n = nc[++ii];
  xx = ws[ii];
  
  // Last recursion level ?	
  //
  if (n == N2) {
    
    // Form even function
    //
    xx[0] = add_cc(data_a[0],data_a[8]);
    xx[1] = add_cc(data_a[1],data_a[7]);
    xx[2] = add_cc(data_a[2],data_a[6]);
    xx[3] = add_cc(data_a[3],data_a[5]);
    
    // Save even DCT coefficients
    //
    t0 = sub_cc(xx[0],data_a[4]);
    t3 = add_cc(xx[0],data_a[4]);
    t2 = add_cc(xx[1],xx[3]);
    t1 = mult_cc(sub_cc(xx[1],xx[3]),sqrt2);
    t4 = add_cc(t3,xx[2]);
      
    coeff_a[0] = add_cc(t4,t2);
    coeff_a[2] = add_cc(t0,t1);
    coeff_a[4] = sub_cc(t3,xx[2]);
    coeff_a[6] = sub_cc(t0,t1);
    coeff_a[8] = sub_cc(t4,t2);
    
    // Form twiddled odd function
    //
    xx[0] = mult_cc(sub_cc(data_a[0],data_a[8]),sec0);
    xx[1] = mult_cc(sub_cc(data_a[1],data_a[7]),sec1);
    xx[2] = mult_cc(sub_cc(data_a[2],data_a[6]),sec2);
    xx[3] = mult_cc(sub_cc(data_a[3],data_a[5]),sec3);
      
    // Save odd DCT coefficients
    //
    t1 = mult_cc(sub_cc(xx[1],xx[3]),sqrt2);
    t2 = add_cc(xx[1],xx[3]);
    t4 = add_cc(xx[0],xx[2]);
      
    coeff_a[1] = add_cc(t4,add_cc(t2,(t0 = add_cc(xx[0],t1))));
    coeff_a[3] = add_cc(t0,(t3 = sub_cc(xx[0],xx[2])));
    coeff_a[5] = add_cc(t3,(t1 = sub_cc(xx[0],t1)));
    coeff_a[7] = sub_cc(add_cc(t1,t4),t2); 
  }
  
  else {
    n2 = shift_cc(FT_RIGHT,n,1);
    n2p1 = add_cc(n2,1);
    nq = ic[ii];  
      
    if ((k = mc[ii]) > n2)
      k = n2;
    
    // Form even-odd functions
    //
    if (nn <= n2p1)
      for (i = j = 0; i < nn; i = add_cc(i,1), j = add_cc(j,nq))
	xx[add_cc(i,n2p1)] = mult_cc((xx[i] = data_a[i]),sc[j]);
    else {
      for (i = j = 0; i <= sub_cc(n,nn); i = add_cc(i,1), j = add_cc(j,nq))
	xx[add_cc(i,n2p1)] = mult_cc((xx[i] =data_a[i]),sc[j]);
      
      for (; i < n2; i = add_cc(i,1), j = add_cc(j,nq)) {
	xx[i] = add_cc(data_a[i],data_a[sub_cc(n,i)]);
	xx[add_cc(i,n2p1)] = mult_cc(sub_cc(data_a[i],
					    data_a[sub_cc(n,i)]),sc[j]);
      }
      
      xx[i] = data_a[i];
      xx[add_cc(i,n2p1)] = 0.0;
    }
    
    // Do recursive DCTs
    //
    real_dct_cc((XX = &xx[add_cc(n,2)]), xx);
    real_dct_cc(&XX[n2p1], &xx[n2p1]);
    
    // Save DCT coefficients
    //
    t0 = XX[n2p1];
      
      for (i = 0; i < k; i = add_cc(i,2)) {
	coeff_a[j = add_cc(i,i)] = XX[i];
	coeff_a[add_cc(j,1)] = add_cc(t0,(t1=XX[add_cc(add_cc(i,n2p1),1)]));
	coeff_a[add_cc(j,2)] = XX[add_cc(i,1)];
	coeff_a[add_cc(j,3)] = add_cc(t1,(t0=XX[add_cc(add_cc(i,n2p1),2)]));
      }
      
      coeff_a[add_cc(i,i)] = XX[i];
    }
  
  // Decrement level & exit
  //
  ii = sub_cc(ii,1);

  // exit gracefully
  //
  return ISIP_TRUE;
  
}

//
// This code is based on the algorithm described in
// H.Guo et.al,"The Quick Discrete Fourier Transform", IEEE Transactions
// on Acoustics Speech and Signal Processing 1994.pp. 445-448
// This code is adapted form the code sent by Sitton of Rice University
//
// arguments:
//  float_8* coeff_a : output dst coefficient array
//  float_8* data_a  : input data array
//
// return: a logical_1 value indicating status
//
// implements a disrete sine transform

logical_1 Fourier_transform::real_dst_cc(float_8* coeff_a, float_8* data_a){
  
  float_8  t0, t1, t2;
  int_4	i, j, k, n, nq, n2, n2m1;

  float_8* XX;
  float_8* xx ;
  
  // Increment level & start	
  //
  n = nc[++ii];
  xx = ws[ii];
  
  // Last recursion level ?	
  //
  if (n == N2){
      // Form odd function
      //
      xx[0] = sub_cc(data_a[0],data_a[6]);
      xx[1] = sub_cc(data_a[1],data_a[5]);
      xx[2] = sub_cc(data_a[2],data_a[4]);
      
      // Save odd DST coefficients
      //
      t0 = mult_cc(add_cc(xx[0],xx[2]),sqrt2);
      
      coeff_a[1] = add_cc(t0,xx[1]);
      coeff_a[3] = sub_cc(xx[0],xx[2]);
      coeff_a[5] = sub_cc(t0,xx[1]);

      // Form twiddled even function
      //
      xx[0] = mult_cc(add_cc(data_a[0],data_a[6]),sec1);
      xx[1] = mult_cc(add_cc(data_a[1],data_a[5]),sec2);
      xx[2] = mult_cc(add_cc(data_a[2],data_a[4]),sec3);
      
      // Save even DST coefficients
      //
      t0 = mult_cc(add_cc(xx[0],xx[2]),sqrt2);
      
      coeff_a[0] = add_cc((t1 = add_cc(t0,xx[1])),data_a[3]);
      coeff_a[2] = sub_cc(add_cc((t2 = sub_cc(xx[0],xx[2])),t1),data_a[3]);
      coeff_a[4] = add_cc(add_cc((t0 = sub_cc(t0,xx[1])),t2),data_a[3]);
      coeff_a[6] = sub_cc(t0,data_a[3]);
    }
  else {
      n2 = shift_cc(FT_RIGHT,n,1);
      n2m1 = sub_cc(n2,1);
      nq = ic[ii];
      
      if ((k = mc[ii]) > n2m1)
        k = n2m1;
      
      // Form even-odd functions
      //
      if (nn <= n2)
        for (i = 0, j = nq; i < sub_cc(nn,1);
	     i = add_cc(i,1),j = add_cc(j,nq))
          xx[add_cc(i,n2m1)] = mult_cc((xx[i] = data_a[i]),sc[j]);
      else  {
	for (i = 0, j = nq; i < sub_cc(n,nn);
	     i = add_cc(i,1), j = add_cc(j,nq))
	  xx[add_cc(i,n2m1)] = mult_cc((xx[i] = data_a[i]),sc[j]);
	
	for (;i < n2m1; i = add_cc(i,1), j = add_cc(j,nq)){
	  xx[i] = sub_cc(data_a[i],data_a[sub_cc(n,add_cc(2,i))]);
	  xx[i+n2m1] = mult_cc(add_cc(data_a[i],
				      data_a[sub_cc(n,add_cc(2,i))]),sc[j]);
	}
      }
      
      // Do recursive DSTs
      //
      real_dst_cc((XX = &xx[sub_cc(n,2)]), xx);
      real_dst_cc(&XX[n2m1], &xx[n2m1]);
      
      // Save DST coefficients
      //
      coeff_a[1] = XX[0];
      
      if (n2 >= nn){
          t0 = coeff_a[0] = XX[n2m1];
          
          for (i = 1; i < k; i = add_cc(i,2)) {
	    coeff_a[j = add_cc(i,i)] = add_cc((t1 = XX[add_cc(i,n2m1)]),t0);
	    coeff_a[add_cc(j,1)] = XX[i];
	    coeff_a[add_cc(j,2)] = add_cc((t0 = XX[add_cc(add_cc(i,n2m1),1)]),t1);
	    coeff_a[add_cc(j,3)] = XX[add_cc(i,1)];
	  }
	  coeff_a[add_cc(i,i)] = t0;
        }
      else {
          coeff_a[0] = add_cc((t1 = XX[n2m1]),(t0 = data_a[n2m1]));
          
          for (i = 1; i < k; i = add_cc(i,2)){
              coeff_a[j = add_cc(i,i)] =
		sub_cc(add_cc((t2 = XX[add_cc(i,n2m1)]),t1), t0);
              coeff_a[add_cc(j,1)] = XX[i];
              coeff_a[add_cc(j,2)] = add_cc(add_cc((t1 = XX[add_cc(add_cc(i,n2m1),1)]),t2),t0);
              coeff_a[add_cc(j,3)] = XX[add_cc(i,1)];
            }
          
          coeff_a[add_cc(i,i)] = sub_cc(t1,t0);
        }
    }
  
  // Decrement level & exit
  //
  ii = sub_cc(ii,1);

  // exit gracefully
  //
  return ISIP_TRUE;
  
}

//  Program: qftinit, Quick Fourier Transform	
//  secant table and workspace initializer.		

logical_1  Fourier_transform::qft_init_cc(){

  int_4	n2, i, m;
  float_8 q;
  
  // Compute some constants	
  //
  m = 0;
  i = N_d;
  while((i = shift_cc(FT_RIGHT,i,1)))
    m = add_cc(m,1);
  
  n2 = shift_cc(FT_RIGHT,N_d,1);
  
  // Allocate the QFT memory	
  //
  if ((ws = new float_8*[m + 3*m + 3*(N_d + m)]) == NULL) {
    return ISIP_FALSE;
  }
  
  nc = (int_4*)&ws[m];
  ic = &nc[m];
  mc = &ic[m];
  sc = (float_8*)&mc[m];
  ws[0] = &sc[sub_cc(n2,1)];
  
  // Initialize tables	
  //
  nc[0] = N_d;
  ic[0] = 1;
  mc[0] = m;
  
  for (i = 1; i < m; i = add_cc(i,1)) {
    nc[i] = shift_cc(FT_RIGHT,nc[sub_cc(i,1)], 1);
    ic[i] = shift_cc(FT_LEFT,ic[sub_cc(i,1)],1);
    
    if (i > 1)
      ws[i] = add_cc(ws[sub_cc(i,1)],
		     mult_cc(2,(add_cc(nc[sub_cc(i,1)],2))));
    else
      ws[i] = add_cc(add_cc(ws[sub_cc(i,1)],nc[i]),1);	
  }	
  
  // Compute scaled secants	
  //
  q = M_PI / N_d;
  
  for (i = 0; i < n2-1; i = add_cc(i,1))
    sc[i] = 0.5 / (float_8)cos(mult_cc(q,(float)i));	

  // exit gracefully
  //
  return ISIP_TRUE;
}


logical_1 Fourier_transform::qft_close_cc(){
  
  // Deallocate memory space
  //
  delete [] ws;
    
  // exit gracefully
  //
  return ISIP_TRUE;
}
