// file: $PDSP/class/fourier_transform/v1.1/fourier_transform.h
//

// this is the header for the fourier_transform class
//

// make sure definitions are only made once
//
#ifndef __FOURIER_TRANSFORM
#define __FOURIER_TRANSFORM

// isip include files
//
#ifndef __ISIP_INTEGRAL
#include <integral.h>
#endif

#ifndef __ISIP_FLOAT_4
#include <float_4.h>
#endif

// forward class declaration
//
class Float_4;

// Fourier_transform: a class that computes the fourier transform of
// some given data.
//
class Fourier_transform {
  
  //---------------------------------------------------------------------------
  //
  // protected data
  //
  //---------------------------------------------------------------------------
protected:
  
  // parameters related to the algorithm specification
  //
  int_4 N_d;					// order
  int_4 algorithm_d;				// numerical identifier for the
					        // algorithm

  // parameters related to the signal data type
  //
  int_4 data_type_d;				// real or complex
  
  // parameters related to timing analysis
  //
  int_4 num_float_mults_d;			// float multiplication counter
  int_4 num_float_adds_d;			// float addition counter
  int_4 num_int_mults_d;			// int multiplication counter
  int_4 num_int_adds_d;			        // int addition counter
  float_4 cpu_time_d;				// cpu time
  int_4 memory_usage_d;                         // memory usage
  int_4 num_shifts_d;                           // number ofinteger shifts
  
  // debugging parameters
  //
  int_4 debug_level_d;				// amount of debugging info

  //---------------------------------------------------------------------------
  //
  // public methods
  //
  //---------------------------------------------------------------------------
public:

  // required methods
  //
  char_1* name_cc();
  int_4 error_handler_cc(char_1* method_name, char_1* message);
  
  // destructors/constructors
  //
  ~Fourier_transform();  
  Fourier_transform();

  // set signal processing parameters
  //
  logical_1 set_cc(int_4 order);
  logical_1 set_cc(int_4 order, int_4 algorithm);  
  logical_1 set_cc(int_4 order, int_4 algorithm, int_4 data_type);

  // get the name of chosen algorithm
  //
  logical_1 get_algorithm_cc(int_4& algorithm);
  logical_1 get_algorithm_name_cc(char_1* algorithm_name);
  
  // computation methods
  //
  logical_1 compute_cc(float_4* output, float_4* input);

  // statistics methods
  //
  logical_1 reset_stats_cc();
  logical_1 get_stats_cc(int_4& num_float_mults, int_4& num_float_adds,
			 int_4& num_int_mults, int_4& num_int_adds,
			 float_4& cpu_time, int_4& memory_usage,
			 int_4& num_shifts);  
  logical_1 display_stats_cc(FILE* fp, int_4 num_iter, int_4 stats);  

  // timing methods
  //
  logical_1 set_cputime_cc(float_4 cpu_time);
  float_4 get_cputime_cc();

  // debugging methods
  //
  logical_1 set_debug_cc(int_4 debug_level);
  logical_1 debug_cc(FILE* fp, char_1* message);
  
  //---------------------------------------------------------------------------
  //
  // private methods
  //
  //---------------------------------------------------------------------------
private:

  // class maintainance methods
  //

  // methods to display varying levels of statistics
  //
  logical_1 display_stats_cpu_cc(FILE* fp, int_4 num_iter);
  logical_1 display_stats_oper_cc(FILE* fp, int_4 num_iter);
  logical_1 display_stats_mem_cc(FILE* fp, int_4 num_iter);  

  // method to check the order of the algorithm and assign a default
  // algorithm if the chosen algorithm does not support the provided
  // order
  //
  logical_1 algorithm_check_cc();
  
  // basic math operations (used to get statistical information)
  //
  float_4 add_cc(float_4 x, float_4 y);
  float_4* add_cc(float_4* x, int_4 y);
  Float_4* add_cc(Float_4* x, int_4 y);
  float_4 sub_cc(float_4 x, float_4 y);
  float_4 mult_cc(float_4 x, float_4 y);
  int_4 add_cc(int_4 x, int_4 y);
  int_4 sub_cc(int_4 x, int_4 y);
  int_4 mult_cc(int_4 x, int_4 y);
  float_4 mult_sum_cc(float_4 x, float_4 y, float_4 sum);
  int_4 shift_cc(int_4 direction,int_4 x,int_4 shift);
  // algorithm methods for real signals
  //
  logical_1 dft_real_cc(float_4* output, float_4* input);    // real dft
  logical_1 fft2_real_cc(float_4* output, float_4* input);   // real radix-2
  logical_1 fft4_real_cc(float_4* output, float_4* input);   // real radix-4
  logical_1 fht_real_cc(float_4* output, float_4* input);    // real fht
  logical_1 srfft_real_cc(float_4* output, float_4* input);  // real srfft
  logical_1 qft_real_cc(float_4* output, float_4* input);    // real qft
  logical_1 ditf_real_cc(float_4* output, float_4* input);   // real ditf

  // algorithm methods for complex signals
  //
  logical_1 dft_complex_cc(float_4* output, float_4* input); // complex dft
  logical_1 fft2_complex_cc(float_4* output,
			    float_4* input);                 // complex radix-2
  logical_1 fft4_complex_cc(float_4* output,
			    float_4* input);		     // complex radix-4
  logical_1 fht_complex_cc(float_4* output, float_4* input); // complex fht
  logical_1 srfft_complex_cc(float_4* output,
			     float_4* input);		     // complex srfft
  logical_1 qft_complex_cc(float_4* output, float_4* input); // complex qft
  logical_1 ditf_complex_cc(float_4* output,float_4* input); // complex ditf
  logical_1 pfa_complex_cc(float_4* output, float_4* input); // complex pfa
    
  // auxillary methods needed for particular algorithms
  //

  // related to ditf
  //
  logical_1 ditf_dit_cc(float_4* input,int_4 v,int_4 gamma,
			int_4 n,float_4* real_result_ptr,
			float_4* imag_result_ptr);
  logical_1 ditf_calc_cc(float_4* input,int_4 v, int_4 gamma,
			 int_4 n,float_4* real_result_ptr,
			 float_4* imag_result_ptr);
  logical_1 ditf_init_cc();
  logical_1 ditf_cleanup_cc();
  
  int_4 powof2_cc();   // computes number of fft stages required

  // related to qft
  //
  logical_1 qft_init_cc();
  logical_1 qft_close_cc();
  logical_1 real_dct_cc(float_4 *coeff, float_4 *data);
  logical_1 real_dst_cc(float_4 *coeff, float_4 *data);

  // lookup table generators 
  //
  logical_1 lookup_cc(float_4 *wreal, float_4 *wimag);
  logical_1 srfft_lookup_cc(float_4* cos_a, float_4* sin_a); 
  logical_1 srfft_new_lookup_cc(float_4* cos_a,	float_4* sin_a);
  
  
};

// end of file
// 
#endif

