// file: $PDSP/class/fourier_transform/v1.1/ft_fft2_0.cc
//

// isip include files
//
#include "fourier_transform.h"
#include <float_4.h>
#include "fourier_transform_constants.h"

// method: fft2_real_cc
//
// this method performs a radix-2 cooley-tukey decimation-in-frequency
// algorithm.  the code follows the exact structure of that in the book by
// proakis and manolakis " digital signal processing " pp. 731-732
//
// arguments:
// float_4* output_a :(output) output data array 
//			length of output array will always be 2*N
//			memory is allocated internally, not by the calling
//			program.
// float_4* input_a : (input) input data array
//			length of input array is N
//			input data memory should be allocated by the
//			calling program.
//
// return: a logical_1 value indicating status
//
// implements a real discrete fourier transform
//
logical_1 Fourier_transform::fft2_real_cc(float_4* output_a,
					  float_4* input_a){
  
  // declare local variables
  //
  int_4 m,l,i1,i2;
  int_4 n1,n2;
  Float_4 real_twid,imag_twid;
  Float_4 tempr,tempi; 
  Float_4 temp_input[N_d];
  
  // get the lookup values
  //
  Float_4 *wr = new Float_4[N_d];
  Float_4 *wi = new Float_4[N_d];

  // create lookup table
  //
  lookup_cc((float_4*)wr,(float_4*)wi);

  // copy input data to temporary memory so that the input data is retained
  // after calculations.
  //
  for (int_4 kk = 0; kk < N_d; kk++) {
    output_a[kk] = input_a[kk];
    temp_input[kk] = 0;
  }
  
  m = powof2_cc();
  n2 = N_d;
  
  for(int_4 i = 0; i < m; i = add_cc(i,1)){
    n1 = n2;
    n2 = shift_cc(FT_RIGHT,n2,1);
    i1 = 0;
    i2 = (int_4)((N_d)/n1);
    
    for(int_4 j = 0; j < n2; j = add_cc(j,1)){
      real_twid = wr[i1];
      imag_twid = wi[i1];
      i1 = add_cc(i1,i2);
      
      for(int_4 k = j; k < N_d; k = add_cc(k,n1)){
	l = add_cc(k,n2);
	tempr = sub_cc(output_a[k],output_a[l]);
	output_a[k] = add_cc(output_a[k],output_a[l]);
	tempi = sub_cc(temp_input[k],temp_input[l]);
	temp_input[k] = add_cc(temp_input[k],temp_input[l]);
	output_a[l] = add_cc(mult_cc(real_twid,tempr),
			     mult_cc(imag_twid,tempi));
	temp_input[l] =  sub_cc(mult_cc(real_twid,tempi),
			     mult_cc(imag_twid,tempr));
      }
    }
  }

  // procedure for bit reversal
  //
  Float_4 temp = 0;
  int j = 0;
  n1 = sub_cc(N_d,1);
  n2 = shift_cc(FT_RIGHT,N_d,1);
  
  for (int_4 i = 0; i < n1 ; ++i){
    if (i < j)
      {
	temp = output_a[j];
	output_a[j] = output_a[i];
	output_a[i] = temp;
	temp = temp_input[j];
	temp_input[j] = temp_input[i];
	temp_input[i] = temp;
      }
    int_4 k = n2;
    while (k <= j)
      {
	j = sub_cc(j,k);
	k = shift_cc(FT_RIGHT,k,1);
      }
    j = add_cc(j,k);
  }

  // interlace the output and store in the output_a array
  //
  for (int_4 kk = N_d - 1; kk >= 0; kk = sub_cc(kk,1)) {
    output_a[2*kk+1] = temp_input[kk];
    output_a[2*kk] = output_a[kk];    
  }

  // set the memory used
  //
  memory_usage_d = temp.get_ref_count_cc();
    
  // delete memory
  //
  delete [] wr;
  delete [] wi;
  
  // exit gracefully
  //
  return ISIP_TRUE;
}
