// file: $PDSP/class/fourier_transform/v1.1/ft_fht_0.cc
//

// system include files
//
#include <math.h>

// isip include files
//
#include "fourier_transform.h"
#include <float_4.h>
#include "ft_fht_0.h"
#include "fourier_transform_constants.h"

// definition required to initialize some functions and arrays specific
// to the fht provided by ron mayer
//
#define FT_FHT_GOOD_TRIG

// method: fht_real_cc
//
// this code is based on the algorithm as described in the work by
// ron mayer whose code is available as a free-ware. the algorithm is
// described in the paper hsieh s. hou,"the fast hartley transform algorithm",
// ieee transactions on computers, february 1987.pp. 147-153
//
//
// arguments:
// float_4* output_a :(output) output data array 
//			length of output array will always be 2*N
//			memory is allocated internally, not by the calling
//			program.
// float_4* input_a : (input) input data array
//			length of input array is N
//			input data memory should be allocated by the
//			calling program.
//
// return: a logical_1 value indicating status
//
// implements a real fast hartley transform
//
logical_1 Fourier_transform::fht_real_cc(float_4* output_a,float_4* input_a){
  
  // declare local variables
  //
  Float_4 a,b;
  Float_4 c1,s1,s2,c2,s3,c3,s4,c4;
  Float_4 f0,g0,f1,g1,f2,g2,f3,g3;
  int_4 i,j,k,k1,k2,k3,k4,kx;
  Float_4* fi,*fn,*gi;
  Float_4 temp_input[N_d];

  // these arrays are defined to simulate the memory defined in the header
  //
  Float_4 temp_sintab[20];
  Float_4 temp_castab[20];
  
  TRIG_VARS;
  
  // copy input data to  temporary memory and zero the output so that
  // the input data is retained after calculations.
  //
  for (int_4 kk = 0; kk < N_d; kk++) {
    temp_input[kk] = input_a[kk];
    output_a[kk] = 0;
  }
  
  for ( k1 = 1, k2 = 0; k1 < N_d; k1 = add_cc(k1,1)){
    for (k=shift_cc(FT_RIGHT,N_d,1); (!((k2^=k)&k)); k=shift_cc(FT_RIGHT,k,1));
    if (k1>k2){
      a=temp_input[k1];
      temp_input[k1]=temp_input[k2];
      temp_input[k2]=a;
    }
  }
  
  for ( k=0 ; (shift_cc(FT_LEFT,1,k)) < N_d ; k = add_cc(k,1)){};
  k  &= 1;
  if (k==0){
    for (fi = temp_input,fn = add_cc(temp_input,N_d); fi < fn; fi = add_cc(fi,4)){
      f1 = sub_cc(fi[0 ],fi[1 ]);
      f0 = add_cc(fi[0 ],fi[1 ]);
      f3 = sub_cc(fi[2 ],fi[3 ]);
      f2 = add_cc(fi[2 ],fi[3 ]);
      fi[2] = sub_cc(f0,f2);	
      fi[0] = add_cc(f0,f2);
      fi[3] = sub_cc(f1,f3);	
      fi[1] = add_cc(f1,f3);
    }
  }

  else {
    
    for (fi = temp_input,fn = add_cc(temp_input,N_d), gi = add_cc(fi,1);
	 fi < fn; fi = add_cc(fi,8), gi = add_cc(gi,8)){
      c1 = sub_cc(fi[0 ],gi[0 ]);
      s1 = add_cc(fi[0 ],gi[0 ]);
      c2 = sub_cc(fi[2 ],gi[2 ]);
      s2 = add_cc(fi[2 ],gi[2 ]);
      c3 = sub_cc(fi[4 ],gi[4 ]);
      s3 = add_cc(fi[4 ],gi[4 ]);
      c4 = sub_cc(fi[6 ],gi[6 ]);
      s4 = add_cc(fi[6 ],gi[6 ]);
      f1 = sub_cc(s1,s2);	
      f0 = add_cc(s1,s2);
      g1 = sub_cc(c1,c2);	
      g0 = add_cc(c1,c2);
      f3 = sub_cc(s3,s4);	
      f2 = add_cc(s3,s4);
      g3 = mult_cc(M_SQRT2,c4);		
      g2 = mult_cc(M_SQRT2,c3);
      fi[4 ] = sub_cc(f0,f2);
      fi[0 ] = add_cc(f0,f2);
      fi[6 ] = sub_cc(f1,f3);
      fi[2 ] = add_cc(f1,f3);
      gi[4 ] = sub_cc(g0,g2);
      gi[0 ] = add_cc(g0,g2);
      gi[6 ] = sub_cc(g1,g3);
      gi[2 ] = add_cc(g1,g3);
    }
  }
  
  if (N_d<16){ 

    if ( algorithm_d == FT_FHT_REAL ){

      //  interlace the output and put into output_a
      //
      output_a[0] = temp_input[0];
      output_a[1] = 0.0;

      for ( int_4 kk = 1; kk <= N_d/2; kk++ ){
	output_a[2*kk] = output_a[2*N_d - 2*kk ] = temp_input[kk];
	output_a[2*N_d - 2*kk + 1 ] = temp_input[N_d - kk];
	output_a[2*kk+1] = -temp_input[N_d - kk];
      }
      output_a[N_d+1] = 0.0;
      
    }

    // set the memory used
    //
    memory_usage_d = a.get_ref_count_cc();

    if ( algorithm_d == FT_FHT_COMPLEX ) {

      // get the output data
      //
      for (int_4 kk = 0; kk < N_d; kk++) {
	output_a[kk] = temp_input[kk];
      }
    }
    
    // exit gracefully
    //
    return ISIP_TRUE;
  }

  do {
    k  = add_cc(k,2);
    k1 = shift_cc(FT_LEFT,1, k);
    k2 = shift_cc(FT_LEFT,k1, 1);
    k4 = shift_cc(FT_LEFT,k2, 1);
    k3  = add_cc(k2,k1);
    kx  = shift_cc(FT_RIGHT,k1, 1);
    fi  = temp_input;
    gi  = add_cc(fi,kx);
    fn  = add_cc(temp_input,N_d);
    do
      {
	f1 = sub_cc(fi[0 ],fi[k1]);
	f0 = add_cc(fi[0 ],fi[k1]);
	f3 = sub_cc(fi[k2],fi[k3]);
	f2 = add_cc(fi[k2],fi[k3]);
	fi[k2] = sub_cc(f0,f2);

	fi[0 ] = add_cc(f0,f2);
	fi[k3] = sub_cc(f1,f3);
	fi[k1] = add_cc(f1,f3);
	g1 = sub_cc(gi[0 ],gi[k1]);
	g0 = add_cc(gi[0 ],gi[k1]);
	g3 = mult_cc(M_SQRT2,gi[k3]);
	g2 = mult_cc(M_SQRT2,gi[k2]);
	gi[k2] = sub_cc(g0,g2);
	gi[0 ] = add_cc(g0,g2);
	gi[k3] = sub_cc(g1,g3);
	gi[k1] = add_cc(g1,g3);
	gi = add_cc(gi,k4);
	fi = add_cc(fi,k4);
      } while (fi < fn);
    
     TRIG_INIT(k,c1,s1);

     for (i = 1; i < kx; i = add_cc(i,1)){
       TRIG_NEXT(k,c1,s1);
       c2 = sub_cc(mult_cc(c1,c1),mult_cc(s1,s1));
       s2 = mult_cc(2,mult_cc(c1,s1));
       fn = add_cc(temp_input,N_d);
       fi = add_cc(temp_input,i);
       gi = add_cc(temp_input, sub_cc( k1,i));
       do {
	 b  = sub_cc(mult_cc(s2,fi[k1]),mult_cc(c2,gi[k1]));
	 a  = add_cc(mult_cc(c2,fi[k1]),mult_cc(s2,gi[k1]));
	 f1 = sub_cc(fi[0 ],a);
	 f0 = add_cc(fi[0 ],a);
	 g1 = sub_cc(gi[0 ],b);
	 g0 = add_cc(gi[0 ],b);
	 b  = sub_cc(mult_cc(s2,fi[k3]),mult_cc(c2,gi[k3]));
	 a  = add_cc(mult_cc(c2,fi[k3]),mult_cc(s2,gi[k3]));
	 f3 = sub_cc(fi[k2],a);
	 f2 = add_cc(fi[k2],a);
	 g3 = sub_cc(gi[k2],b);
	 g2 = add_cc(gi[k2],b);
	 b  = sub_cc(mult_cc(s1,f2),mult_cc(c1,g3));
	 a  = add_cc(mult_cc(c1,f2),mult_cc(s1,g3));
	 fi[k2]  = sub_cc(f0,a);
	 fi[0 ]  = add_cc(f0,a);
	 gi[k3]  = sub_cc(g1,b);
	 gi[k1]  = add_cc(g1,b);
	 b  = sub_cc(mult_cc(c1,g2),mult_cc(s1,f3));
	 a  = add_cc(mult_cc(s1,g2),mult_cc(c1,f3));
	 gi[k2]  = sub_cc(g0,a);
	 gi[0 ]  = add_cc(g0,a);
	 fi[k3]  = sub_cc(f1,b);
	 fi[k1]  = add_cc(f1,b);
	 gi = add_cc(gi,k4);
	 fi = add_cc(fi,k4);
       } while (fi < fn);
     }
     TRIG_RESET(k,c1,s1);
   } while (k4 < N_d);
  
  a = b = 0;  

  if (algorithm_d == FT_FHT_COMPLEX){
    
    // get the output data
    //
    for (int_4 kk = 0; kk < N_d; kk++) {
      output_a[kk] = temp_input[kk];
    }
  }
    
  
  if (algorithm_d == FT_FHT_REAL){

    // get fourier coefficients from hartley coefficients
    //
    for (i = 1,j = sub_cc(N_d,1), k = (N_d/2); i < k; i = add_cc(i,1),j = sub_cc(j,1)) {
      a = temp_input[i];
      b = temp_input[j];
      temp_input[j] = mult_cc(sub_cc(a,b),0.5);
      temp_input[i] = mult_cc(add_cc(a,b),0.5);
    }
   
    //  interlace the output and put into output_a
    //
    output_a[0] = temp_input[0];
    output_a[1] = 0;
    
    for ( int_4 kk = 1; kk <= N_d/2; kk = add_cc(kk,1)){
      output_a[2*kk] = output_a[2*N_d - 2*kk ] = temp_input[kk];
      output_a[2*kk+1] = -temp_input[N_d - kk];
      output_a[2*N_d - 2*kk + 1 ] = temp_input[N_d - kk];
    }
    output_a[N_d+1] = 0.0;
    
  }

  // set the memory used
  //
  memory_usage_d = a.get_ref_count_cc();
    
  // exit gracefully
  //
  return ISIP_TRUE;
}





