// file: $PDSP/class/fourier_transform/v2.0/fourier_transform.h
//
// this file defines the freq_response  class
//

// make sure definitions are only made once
//
#ifndef __FOURIER_TRANSFORM
#define __FOURIER_TRANSFORM

// isip include files
//
#ifndef __ISIP_INTEGRAL
#include <math/integral/integral.h>
#endif

// include system files
//
#include <time.h>
#include <string.h>

// Fourier_transform: a class that computes the fourier transform of
// some given data.
//
class Fourier_transform {
  
  //---------------------------------------------------------------------------
  //
  // protected data
  //
  //---------------------------------------------------------------------------
protected:
  
  // parameters related to the algorithm specification
  //
  int_4 N_d;					// order
  char_1* algorithm_name_d;			// specific algorithm
  int_4 algorithm_d;

  // parameters related to the signal data type
  //
  int_4 data_type_d;				// real or complex
  int_4 transform_direction_d;			// forward or reverse
  
  // parameters related to timing analysis
  //
  int_4 num_float_mults_d;			// float multiplication counter
  int_4 num_float_adds_d;			// float addition counter
  int_4 num_int_mults_d;			// int multiplication counter
  int_4 num_int_adds_d;			        // int addition counter
  float_8 cpu_time_d;				// cpu time
  int_4 num_iter_d;				// number of iterations
  int_4 memory_usage_d;                         // memory usage
  int_4 num_shifts_d;                           // number ofinteger shifts
  
  // debugging parameters
  //
  int_4 debug_level_d;				// amount of debugging info
  int_4 stats_d;                                // amount of statistics

  //---------------------------------------------------------------------------
  //
  // public methods
  //
  //---------------------------------------------------------------------------
public:

  // required methods
  //
  char_1* name_cc();
  int_4 error_handler_cc(char_1* method_name, char_1* message);
  
  // destructors/constructors
  //
  ~Fourier_transform();  
  Fourier_transform();

  // algorithm specification
  //
  logical_1 set_cc(int_4 order_a, int_4 type_a, int_4 num_iter_a,
 		   int_4 dbg_level_a, int_4 stats_a, char_1* name_a);
  logical_1 method_choice_cc();
  
  // computation methods
  //
  logical_1 compute_cc(float_8* output, float_8* input);
  logical_1 compute_cc (float_8* real_coeff_a, float_8* imag_coeff_a,
                        float_8* real_data_a,float_8* imag_data_a);

  // timing methods
  //
  logical_1 reset_cc();
  logical_1 get_statistics_cc(int_4 num_mult, int_4 num_add,
 			      int_4 mem_size, float_8 num_secs);
  logical_1 print_statistics_cc(FILE* fp, int_4 stats_level_a,
 				int_4 num_iter_a);
  logical_1 set_cputime_cc(float_8 cpu_time_a);
  float_8 get_cputime_cc();
  
  // debugging methods
  //
  logical_1 set_debug_cc(int_4 debug_level);
  logical_1 debug_cc(FILE* fp, char_1* message);
  
  //---------------------------------------------------------------------------
  //
  // private methods
  //
  //---------------------------------------------------------------------------
private:

  // algorithm methods for real signals
  //
  logical_1 dft_real_cc(float_8* in_r, float_8* in_i);     // real DFT
  logical_1 fft2_real_cc(float_8* in_r, float_8* in_i);    // real RADIX-2
  logical_1 fft4_real_cc(float_8* in_r, float_8* in_i);    // real RADIX-2
  logical_1 fht_real_cc(float_8* in_r);                    // real FHT
  logical_1 srfft_real_cc(float_8* in_r, float_8* in_i);   // real SRFFT
  logical_1 qft_real_cc(float_8 *real_coeff_a,float_8 *imag_coeff_a,
                                      float_8 *data_a);
  logical_1 ditf_real_cc(float_8 *in_r, float_8* in_i);

  // algorithm methods for complex signals
  //
  logical_1 dft_complex_cc(float_8* in_r, float_8* in_i);  // complex DFT
  logical_1 fft2_complex_cc(float_8* in_r, float_8* in_i); // complex RADIX-2
  logical_1 fft4_complex_cc(float_8* in_r, float_8* in_i); // complex RADIX-4
  logical_1 fht_complex_cc(float_8* in_r, float_8* in_i);  // complex FHT
  logical_1 srfft_complex_cc(float_8* in_r, float_8* in_i);// complex SRFFT
  logical_1 qft_complex_cc(float_8* X, float_8* Y, float_8* x, float_8* y);
  logical_1 ditf_complex_cc(float_8* in_r, float_8* in_i);

  // auxillary methods need for particular algorithms
  //
  logical_1 Fourier_transform::ditf_dit_cc(float_8* input,int_4 v,int_4 gamma,
					   int_4 two_v,int_4 n,
					   float_8* real_result_ptr,
					   float_8* imag_result_ptr);
  
  logical_1 Fourier_transform::ditf_calc_cc(float_8* input,int_4 v,
					    int_4 gamma,int_4 two_v,int_4 n,
					    float_8* real_result_ptr,
					    float_8* imag_result_ptr,
					    int_4 end_stage);



  int_4 powof2_cc();   // computes number of FFT stages required
  logical_1 qft_init_cc();
  logical_1 qft_close_cc();
  logical_1 real_dct_cc(float_8 *coeff_a, float_8 *data_a);
  logical_1 real_dst_cc(float_8 *coeff_a, float_8 *data_a);

  // lookup table generators 
  //
  logical_1 lookup_cc(float_8 *wreal, float_8 *wimag);
  logical_1 srfft_lookup_cc(float_8* cos_a, float_8* sin_a); // lookup table
  logical_1 srfft_new_lookup_cc(float_8* cos_a,
 				float_8* sin_a); // lookup table
  
  // basic math operations (used to get statistical information)
  //
  float_8 add_cc(float_8 x, float_8 y);
  float_8* add_cc(float_8* x, int_4 y);
  float_8 sub_cc(float_8 x, float_8 y);
  float_8 mult_cc(float_8 x, float_8 y);
  int_4 add_cc(int_4 x, int_4 y);
  int_4 sub_cc(int_4 x, int_4 y);
  int_4 mult_cc(int_4 x, int_4 y);
  float_8 mult_sum_cc(float_8 x, float_8 y, float_8 sum);
  int_4 shift_cc(int_4 direction,int_4 x,int_4 shift);
};

// end of include file
// 
#endif







