// file: $PDSP/class/fourier_transform/v2.0/ft_ditf_0.cc
//

// include files
//
#include "fourier_transform.h"
#include "fourier_transform_constants.h"
//#include <float_8/v1.0/float_8.h>

// define the core term of W
#define ARG_CORE ISIP_TWOPI/N_d

// DITF externals
//
static int_4* two_v_table;   // holds the powers of 2
static int_4** mod_table;    // holds the mod factors

// method: ditf_real_cc
//
// This method is based on the DITF algorithm by Ali Saidi
// found in the 1994 ICASSP Proceedings 
//
// arguments:
//	float_8* output_a : will hold the imaginary output values
//	float_8* input_a : will hold the real output values
//                         holds the data on input
//
// return: a logical_1 value indicating status
//
logical_1 Fourier_transform::ditf_real_cc(float_8* input_a,
					  float_8* output_a) {

  // declare local variables
  //
  
  float_8 real_a[N_d];
  int_4 v = powof2_cc();
  int_4 i,t;			  // incremental counter index
  int_4 two_v;  		  // two_v = 2^v
  int_4 end_stage = shift_cc(FT_RIGHT,v,1);
  int_4 k;

  //  input_a[0]=0;
  //  input_a[1]=1;
  //  input_a[2]=1;
  //  input_a[3]=0;

  // diagnostic printing
  //
    for ( int kk = 0; kk < N_d; kk++ ){
      input_a[kk] = real_a[kk];
      printf ("%f %f\n", input_a[kk], output_a[kk]);
    }
  
  mod_table = (int_4**)malloc(sizeof(int_4*)*(N_d));
  // create lookup table for mod products;
  //
  for(i=0;i<N_d;i++){
    mod_table[i]=(int_4*)malloc(sizeof(int_4)*(v+1));
    (mod_table[i])[0]=i;
    t=2;
    for(k=1;k<=v;k++){
      (mod_table[i])[k]=(i+N_d/t)%N_d;
      t=t<<1;
    }
  }


  // create lookup table
  //
  two_v = 1;
  two_v_table = (int_4*)malloc(sizeof(int_4)*v);
  for (t=0;t<=v;t++){
    two_v_table[t]=two_v;
    two_v = two_v*2;
  }
  two_v = N_d;
  
  
  // calculate 0..N/2-1 ditf coefficients
  //
  k = (N_d-1);
  for (i=0;i<N_d/2;i++) {
    ditf_calc_cc(input_a,v,i,two_v,i,&real_a[i],&output_a[i],end_stage);
    t = k-i;
    real_a[t] = real_a[i];
    output_a[t] = -output_a[i];
  }
  


  // diagnostic printing
  //
    for ( int kk = 0; kk < N_d; kk++ ){
      input_a[kk] = real_a[kk];
      printf ("%f %f\n", input_a[kk], output_a[kk]);
    }
  // exit gracefully
  //
  return ISIP_TRUE;
  
}


// subroutine to calculate the dft using the recursive decimation-in-time
// algorithm described by Ali Saidi in the ICASSP Proceedings of 1994.
// this DIT should not be used alone because it outputs values in a special
// form to be consistent with the DITF
//
// input: input - array of sample points
//   		 v - stage currently in
//			 gamma - node of the stage
//			 two_v - 2^v
//			 n - sample point currently calculating
//			 N - number of samples
//			 real_result_ptr - pointer to the real part of the
//                            calculated result
//			 imag_result_ptr - pointer to the imag part of the
//                            calculated result
//
// output: None
//
// description: this function uses a recursive call to itself to calculate the
//		 coefficients of the dft at gamma node at stage v for the nth
//		 sample point using the following recursion.
//              if gamma = v = 0
//		       g'[gamma,v](n) = x(n)
//              if gamma < 2^(v-1)
//     	         g'[gamma,v](n) = g'[gamma,v-1](n) +
//		  g'[gamma,v-1]((n+N/(2^v))%N_d)W^(gamma/2^(v-1))
//		else
//		  lambda = gamma - 2^(v-1)
//		  g'[gamma,v](n) = g'[gamma,v-1](n) -
//		  g'[gamma,v-1]((n+N/(2^v))%N_d)W^(lambda/2^(v-1))
//
//
logical_1 Fourier_transform::ditf_dit_cc(float_8* input,int_4 v,int_4 gamma,
					 int_4 two_v,int_4 n,
					 float_8* real_result_ptr,
					 float_8* imag_result_ptr){
  
  
  // define local variables
  //
  float_8 real1 = 0.0,
    real2 = 0.0,
    real3 = 0.0,
    imag1 = 0.0,
    imag2 = 0.0,
    imag3 = 0.0;
  float_8 arg=0;
  float_8 x1=0;
  float_8 x2=0;
  int_4 half_two_v=0;
  
  // define variables that will be used often to avoid repeat calculations
  //
  int_4 v_minus = v-1;
  
  if (v != 0){
    half_two_v = two_v_table[v_minus];
  }
  
  // if v = gamma = 0
  // assign return x[n]
  //
  if ((v == 0) && (gamma == 0)) {
    *real_result_ptr = input[n];
    *imag_result_ptr = (float) 0;
  }
  
  // if gamma <= 2^(v-1) do recursion
  //
  else if (gamma < half_two_v) {
    
    // calculate complex multiplier
    //
    if (gamma==0){
      arg = 0;
      x1 = 1;
      x2 = 0;
    }
    else if(gamma==1){
      arg = ISIP_TWOPI/two_v;
      x1 = cos(arg);
      x2 = sin(arg);
    }
    else {
      arg = mult_cc((float_8)gamma,ISIP_TWOPI)/two_v;
      x1 = cos(arg);
      x2 = sin(arg);
    }
    
    
    // recursively compute the DIT
    //
    ditf_dit_cc(input,v_minus,gamma,half_two_v,n,&real1,&imag1);
    ditf_dit_cc(input,v_minus,gamma,half_two_v,(mod_table[n])[v],&real2,
		&imag2);
    
    // do complex multiplication and store results
    //
    if ((real2==0.0)&&(imag2==0.0)){
      real3=0;
      imag3=0;
    }
    else if ((real2==1)&&(imag2==1)) {
      real3 = add_cc(x1,x2);
      imag3 = sub_cc(x1,x2);
    }
    else if (x1==0) {
      if (x2==1){
	real3 = imag2;
	imag3 = sub_cc(0,real2);
      }
      else {
	real3 = sub_cc(0,imag2);
	imag3 = real2;
      }
    }
    else if(x2==0){
      if(x1==1){
	real3 = real2;
	imag3 = imag2;
      }
      else {
	real3 = -real2;
	imag3 = -imag2;
      }
    }
    else {
      real3 = add_cc(mult_cc(real2,x1),mult_cc(imag2,x2));
      imag3 = sub_cc(mult_cc(imag2,x1),mult_cc(real2,x2));
    }
    
    if (real3==0){
      *real_result_ptr = real1;
    }
    else{
      *real_result_ptr = add_cc(real1,real3);
    }
    if (imag3==0){
      *imag_result_ptr=imag1;
    }
    else{
      *imag_result_ptr = add_cc(imag1,imag3);
    }
  }    
  
  
  // if gamma > 2^(v-1) then do recursion
  //
  else {
    
    // assign lambda = gamma - 2^(v-1)
    //
    int lambda = sub_cc(gamma,half_two_v);
    
    // compute complex multiplier
    //
    if (lambda == 0) {
      arg = 0;
      x1 = 1;
      x2 = 0;
    }
    else {
      arg = (mult_cc((float_8)lambda,ISIP_TWOPI) / two_v);
      x1 = cos(arg);
      x2 = sin(arg);
    }
    
    
    // recursively compute DIT
    //
    ditf_dit_cc(input,v_minus,lambda,half_two_v,mod_table[n][0],&real1,&imag1);
    ditf_dit_cc(input,v_minus,lambda,half_two_v,mod_table[n][v],
		&real2,&imag2);
    
    // do complex multiplication and store results
    if ((real2==0.0)&&(imag2==0.0)) {
      real3=0;
      imag3=0;
    }
    else if ((real2==1)&&(imag2==1)) {
      real3=add_cc(x1,x2);
      imag3=sub_cc(x1,x2);
    }
    else if (x1==0) {
      if (x2==1){
	real3 = imag2;
	imag3 = -real2;
      }
      else {
	real3 = -imag2;
	imag3 = real2;
      }
    }
    else if (x2==0) {
      if (x1==1){
	real3 = real2;
	imag3 = imag2;
      }
      else {
	real3 = -real2;
	imag3 = -imag2;
      }
    }
    
    else {
      real3 = add_cc(mult_cc(real2,x1),mult_cc(imag2,x2));
      imag3 = sub_cc(mult_cc(imag2,x1),mult_cc(real2,x2));
    }
    
    if (real3==0){
      *real_result_ptr = real1;
    }
    else{
      *real_result_ptr = sub_cc(real1,real3);
    }
    if (imag3==0){
      *imag_result_ptr=imag1;
    }
    else{
      *imag_result_ptr = sub_cc(imag1,imag3);
    }
  }
  
  
  return ISIP_TRUE;
}



// subroutine to calculate the dft using the decimation-in-time-frequency
// algorithm described by Ali Saidi in the ICASSP Proceedings of 1994.
//
// input: input - array of sample points
//   		 v - stage currently in
//			 gamma - node of the stage
//			 two_v - 2^v
//			 n - sample point currently calculating
//			 N - number of samples
//			 real_result_ptr - pointer to the real part of the
//                           calculated result
//			 imag_result_ptr - pointer to the imag part of the
//                           calculated result
//			 end_stage - the stage at which to transition to the
//                           recursive DIT
//
// output: None
//
// description: this function uses a recursive call to itself to calculate the
// 	 coefficients of the dft at the gamma node of stage v for the nth
//	 sample point using the following recursion.
//         if gamma = v = 0
//		g[gamma,v](n) = x(n)
//         if gamma < 2^(v-1)
//		g[gamma,v](n) = g[gamma,v-1](n) +
//		 g[gamma,v-1]((n+N/(2^v))%N)
//	 else
//        	 lambda = gamma - 2^(v-1)
//	         g[gamma,v](n) = (g[gamma,v-1](n) -
//		  g[gamma,v-1]((n+N/(2^v))%N))W^(n*2^(v-1))
//
//	this is repeated until v = (log2(N))/2
//	at that point the remaining iterations are carried out by the
//	recursive decimation-in-time algorithm and the result is
//	multiplied by W^(gamma*n)
//
logical_1 Fourier_transform::ditf_calc_cc(float_8* input,int_4 v,
					  int_4 gamma,int_4 two_v,int_4 n,
					  float_8* real_result_ptr,
					  float_8* imag_result_ptr,
					  int_4 end_stage){
  
  
  // local variables
  //
  
  float_8 real1 = 0.0,
    real2 = 0.0,
    imag1 = 0.0,
    imag2 = 0.0;
  float_8 x1,x2;
  float_8 arg;
  int_4 half_two_v=0;
  
  // determine if it is time to switch to the DIT portion of the recursion
  //
  if (v > end_stage) {
    
    //  assign variables which are used multiple times to avoid recalculation
    //
    int v_minus = sub_cc(v,1);
    if (v != 0){
      half_two_v = two_v_table[v_minus];
    }
    
    // if v = gamma = 0, return x[n]
    //
    if ((v == 0) && (gamma == 0)) {
      *real_result_ptr = input[n];
      *imag_result_ptr = (float) 0;
    }
    //**************IMPROVEMENTS**************//    
    // if gamma <= 2^(v-1) do recursion
    //
    else if (gamma < (half_two_v)) {
      ditf_calc_cc(input,v_minus,gamma,half_two_v,mod_table[n][0],&real1,
		   &imag1,end_stage);
      ditf_calc_cc(input,v_minus,gamma,half_two_v,mod_table[n][v],
		   &real2,&imag2,end_stage);
      *real_result_ptr = add_cc(real1,real2);
      *imag_result_ptr = add_cc(imag1,imag2);
    }
    
    
    // if gamma > 2^(v-1) do recursion
    //
    else {
      int lambda = sub_cc(gamma,half_two_v);
      if (n==0) {
	arg = 0;
	x1 = 1;
	x2 = 0;
      }
      else if ((n==1)&&(half_two_v==1)){
	arg = ARG_CORE;
	x1 = cos(arg);
	x2 = sin(arg);
      }
      else if (half_two_v == 1) {
	arg = mult_cc((float_8)n,ARG_CORE);
	x1 = cos(arg);
	x2 = sin(arg);
      }
      else {	
	arg = mult_cc((float_8)mult_cc(n,half_two_v),ARG_CORE);
	x1 = cos(arg);
	x2 = sin(arg);
      }
      ditf_calc_cc(input,v_minus,lambda,half_two_v,mod_table[n][0],&real1,
		   &imag1,end_stage);
      ditf_calc_cc(input,v_minus,lambda,half_two_v,mod_table[n][v],
		   &real2,&imag2,end_stage);
      //**************IMPROVEMENTS***************//
      
      real1 = sub_cc(real1,real2);
      imag1 = sub_cc(imag1,imag2);
      
      // calculate W^(n*2^(v-1))
      //
      
      // do complex multiplication
      //
      if ((real1==0.0)&&(imag1==0.0)) {
	*real_result_ptr = 0;
	*imag_result_ptr = 0;
      }
      else if (x1 == 0) {
	if (x2==1.0) {
	  *real_result_ptr = imag1;
	  *imag_result_ptr = -real1;
	}
	else {
	  *real_result_ptr = -imag1;
	  *imag_result_ptr = real1;
	}
      }
      else if (x2 == 0) {
	if (x1==1.0) {
	  *real_result_ptr = real1;
	  *imag_result_ptr = imag1;
	}
	else {
	  *real_result_ptr = -real1;
	  *imag_result_ptr = -imag1;
	}	
      }
      
      else {
	*real_result_ptr = add_cc(mult_cc(real1,x1),mult_cc(imag1,x2));
	*imag_result_ptr = sub_cc(mult_cc(imag1,x1),mult_cc(real1,x2));
      }
    }
  }
  
  // switch to the DIT portion of the recursion
  //
  else {
    
    // calculate complex conversion factor
    //
    if ((n==0)||(gamma==0)) {
      arg = 0;
      x1 = 1;
      x2 = 0;
    }
    else if ((n==1)&&(gamma==1)) {
      arg = ARG_CORE;
      x1 = cos(arg);
      x2 = sin(arg);
      
    }
    else if (n==1) {
      arg = mult_cc((float_8)gamma,ARG_CORE);
      x1 = cos(arg);
      x2 = sin(arg);
    }
    else if (gamma==1) {
      arg = mult_cc((float_8)n,ARG_CORE);
      x1 = cos(arg);
      x2 = sin(arg);
    }
    else {
      arg = mult_cc((float_8)mult_cc(gamma,n),ARG_CORE);
      x1 = cos(arg);
      x2 = sin(arg);
    }
    
    // call the DIT recursion and multiply by conversion factor
    //
    ditf_dit_cc(input,v,gamma,two_v,mod_table[n][0],&real1,&imag1);
    
    if ((imag1==0.0)&&(real1==0.0)){
      *real_result_ptr = 0;
      *imag_result_ptr = 0;
    }
    else if (x1 == 0) {
      if(x2==1) {
	*real_result_ptr = imag1;
	*imag_result_ptr = -real1;
      }
      else {
	*real_result_ptr = -imag1;
	*imag_result_ptr = real1;
      }	
    }
    else if (x2==0) {
      if(x1==1) {
	*real_result_ptr = real1;
	*imag_result_ptr = imag1;
      }
      else {
	*real_result_ptr = -real1;
	*imag_result_ptr = -imag1;
      }
    }
    else if ((imag1==1)||(real1==1)){
      if ((imag1==1)&&(real1==1)){
	*real_result_ptr = add_cc(x1,x2);
	*imag_result_ptr = sub_cc(x1,x2);
      }
      else if(imag1==1){
	*real_result_ptr = add_cc(mult_cc(real1,x1),x2);
	*imag_result_ptr = sub_cc(x1,mult_cc(real1,x2));
      }
      else if(real1==1){
	*real_result_ptr = add_cc(x1,mult_cc(imag1,x2));
	*imag_result_ptr = sub_cc(mult_cc(imag1,x1),x2);
      }
    }
    
    else {	
      *real_result_ptr = add_cc(mult_cc(real1,x1),mult_cc(imag1,x2));
      *imag_result_ptr = sub_cc(mult_cc(imag1,x1),mult_cc(real1,x2));
    }
    
  }
  // exit gracefully
  //
  return ISIP_TRUE;
  
}



