// file: $PDSP/util/certify/v1.0/certify.cc
//
// this program is intended to certify all algorithms currently available
// under the fourier transform class.
//

// system include files
//
#include <time.h>

// isip include files
//
#include <fourier_transform.h>
#include "certify.h"
#include <integral_constants.h>
#include <fourier_transform_constants.h>
#include "certify_constants.h"

// main program
//
main(int argc, char** argv) {

  // declare a sampling of dft orders and algorithms
  //
  int_4 num_algs = 7;
  int_4 num_orders = 13;
  int_4 orders[] = {2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096,
  		    8192};

  // define default values
  //
  int_4 debug_level = ISIP_DEBUG_NONE;

  // create a fourier transform object
  //
  Fourier_transform ft;

  // get the parameters from the command line
  //
  if (get_params_cc(argc, argv, debug_level) == ISIP_FALSE) {
    return ISIP_ERROR;
  }
  fprintf(stdout, "<---- fourier transform certification --->\n\n");

  if (debug_level != ISIP_DEBUG_NONE) {
    fprintf(stdout, "debug level = %ld\n\n", debug_level);
  }
  
  // initialize performance counters
  //
  float_4 snr[num_algs][num_orders];

  // loop over all algorithms and orders
  //
  for (int_4 nalg=0; nalg<num_algs; nalg++) {
    for (int_4 norder=0; norder<num_orders; norder++) {

      // generate a random number seed
      //
      time_t tloc;
      time(&tloc);
      uint_4 seedval = 2 * ((tloc + 2*norder) / 2) + 1;
      srand(seedval);

      // generate some interesting test data for a real signal
      //
      float_4 sig_r[orders[norder]];
    
      for (int_4 n=0; n<orders[norder]; n++) {
      	sig_r[n] = (float_4)rand() / ISIP_2EXP15;
      }

      if (debug_level > ISIP_DEBUG_NONE) {
	fprintf(stdout, "\n  input signal:\n");
	for (int_4 n=0; n<orders[norder]; n++) {
	  fprintf(stdout, "\t%5ld: %12.4f\n", n, sig_r[n]);
	}
      }

      // declare data:
      // the stategy here is to do two forward transforms to get back the real
      // signal, and then to compute snr in the time domain (cute, huh?)
      //
      float_4* spect = new float_4[2*orders[norder]];
      float_4* sig_c = new float_4[2*orders[norder]];

      
      // do one transform to generate a spectrum
      //
      if ((ft.set_cc(orders[norder], nalg, FT_REAL) == ISIP_FALSE) ||
	  (ft.compute_cc(spect, sig_r) == ISIP_FALSE)) {
	fprintf(stdout,
		"%s: error doing forward fft (alg: %ld order: %ld)\n",
		argv[0], nalg, orders[norder]);
	return ISIP_FALSE;
      }
		
      if (debug_level > ISIP_DEBUG_BRIEF) {
 	fprintf(stdout, "\n  spectrum:\n");
	for (int_4 n=0; n<2*orders[norder]; n++) {
	  fprintf(stdout, "\t%5ld: %12.4f\n", n, spect[n]);
	}
      }

      // compute the second transform - this should net the original signal
      //
      if ((ft.set_cc(orders[norder], nalg, FT_COMPLEX) == ISIP_FALSE) ||
	  (ft.compute_cc(sig_c, spect) == ISIP_FALSE)) {
	fprintf(stdout,
		"%s: error doing forward fft (alg: %ld order: %ld)\n",
		argv[0], nalg, orders[norder]);
	return ISIP_FALSE;
      }

      if (debug_level > ISIP_DEBUG_DETAILED) {
	fprintf(stdout, "\n  new (complex) signal:\n");
	for (int_4 n=0; n<2*orders[norder]; n++) {
	  fprintf(stdout, "\t%5ld: %12.4f\n", n, sig_c[n]);
	}
      }
    
      // circularly shift the data and scale
      //
      float_4 sig_r_new[2*orders[norder]];
      float_4 scale_factor = 1.0 / (float_4)orders[norder];

      sig_r_new[0] = scale_factor * sig_c[0];
      for (int_4 n=1; n<orders[norder]; n++) {
	sig_r_new[n] = scale_factor * sig_c[2*(orders[norder]-n)];
      }

      if (debug_level > ISIP_DEBUG_NONE) {
	fprintf(stdout, "\n  circularly shifted signal:\n");
	for (int_4 n=0; n<orders[norder]; n++) {
	  fprintf(stdout, "\t%5ld: %12.4f\n", n, sig_r_new[n]);
	}
      }
    
      // compute the snr
      //
      float_4 sum_d = 0.0;
      float_4 sum_n = 0.0;
      
      for (int_4 n=0; n<orders[norder]; n++) {
	sum_d += pow(sig_r[n], 2.0);
	sum_n += pow(sig_r[n] - sig_r_new[n], 2.0);
      }
      
      if (sum_n < ISIP_MINIMUM_FLOAT_4) {
	sum_n = ISIP_DB_POW_MIN_VALUE;
      }
      if (sum_d < ISIP_MINIMUM_FLOAT_4) {
	sum_d = ISIP_DB_POW_MIN_VALUE;
      }
      snr[nalg][norder] = ISIP_DB_POW_SCALE_FACTOR * log10(sum_d/sum_n);
      
      // give some status
      //
      char_1 alg_name[ISIP_MAX_STRING_LENGTH];

      if (ft.set_cc(orders[0], nalg, FT_REAL) == ISIP_FALSE) {
	fprintf(stdout, "%s: unknown algorithm (%ld)\n", argv[0], nalg);
	return ISIP_FALSE;
      }
      ft.get_algorithm_name_cc((char_1*)alg_name);
      fprintf(stdout,"alg: %s, order = %6.0ld, snr = %8.2f db\n",
	      alg_name, orders[norder], snr[nalg][norder]);

      // free up space
      //
      delete [] sig_c;
      delete [] spect;
    }
  }
  
  // provide a summary message
  //
  float_4 avg_snr = 0.0;
  int_4 num_ffts = 0;

  for (int_4 nalg=0; nalg<num_algs; nalg++) {
    for (int_4 norder=0; norder<num_orders; norder++) {
      avg_snr += snr[nalg][norder];
      num_ffts++;
    }
  }
  fprintf(stdout, "\naverage snr for %ld ffts was %8.2f db\n",
	  num_ffts, avg_snr/num_ffts);
  
  // exit gracefully
  //
  return ISIP_NO_ERROR;
}
